from enum import Enum
from fuse.data.key_types import DataTypeBasic, TypeDetectorPatternsBased


class DataTypeForTesting(Enum):
    """
    Possible data types stored in sample_dict.
    Using Patterns - the type will be inferred from the key name
    """

    # Default options for types
    IMAGE_FOR_TESTING = (0,)  # Image
    SEG_FOR_TESTING = (1,)  # Segmentation Map
    BBOX_FOR_TESTING = (2,)  # Bounding Box
    CTR_FOR_TESTING = (3,)  # Contour


PATTERNS_DICT_FOR_TESTING = {
    r".*img_for_testing$": DataTypeForTesting.IMAGE_FOR_TESTING,
    r".*seg_for_testing$": DataTypeForTesting.SEG_FOR_TESTING,
    r".*bbox_for_testing$": DataTypeForTesting.BBOX_FOR_TESTING,
    r".*ctr_for_testing$": DataTypeForTesting.CTR_FOR_TESTING,
    r".*$": DataTypeBasic.UNKNOWN,
}

type_detector_for_testing = TypeDetectorPatternsBased(PATTERNS_DICT_FOR_TESTING)



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import copy
import os

import torch
from typing import Sequence, Dict, List


class ModelEnsemble(torch.nn.Module):
    """
    Ensemble Module - runs several sub-modules sequentially.
    In addition to producing a dictionary with predictions of each model in the ensemble,
    the class produces average over the predictions, and majority vote.

    """

    def __init__(self, input_model_dirs: Sequence[str]) -> None:
        super().__init__()
        self.nets = []
        self.input_model_dirs = input_model_dirs

        # Load multiple modules
        for input_model_dir in input_model_dirs:
            net_path = os.path.join(input_model_dir, "net.pth")
            net = torch.load(net_path, map_location="cpu")
            self.nets.append(net)

        # Register modules
        self.nets = torch.nn.ModuleList(self.nets)

    def eval(self) -> None:
        for net in self.nets:
            net.eval()

    def load_state_dict(self, *net_state_dicts: List[Dict], strict: bool) -> None:

        for idx, net_state_dict in enumerate(net_state_dicts):
            self.nets[idx].load_state_dict(net_state_dict, strict=strict)

    def calculate_ensemble_avg(self, ensemble_pred: List[Dict]) -> Dict:
        ensemble_avg_dict = {}
        for key in ensemble_pred[0].keys():
            if "loss" in key:
                continue
            accumulate_prediction = []
            for pred in ensemble_pred:
                accumulate_prediction.append(pred[key])
            accumulate_prediction = torch.stack(accumulate_prediction)
            ensemble_avg_dict[key + "_ensemble_average"] = torch.mean(accumulate_prediction, dim=0)
        return ensemble_avg_dict

    def calculate_ensemble_majority_vote(self, ensemble_pred: List[Dict]) -> Dict:
        ensemble_majority_dict = {}
        for key in ensemble_pred[0].keys():
            if "loss" in key:
                continue
            accumulate_prediction = []
            for pred in ensemble_pred:
                accumulate_prediction.append(torch.argmax(pred[key], dim=-1))
            accumulate_prediction = torch.stack(accumulate_prediction)
            unique_val, unique_count = torch.unique(accumulate_prediction, dim=0, return_counts=True)
            ensemble_majority_dict[key + "_ensemble_majority_vote"] = unique_val[torch.argmax(unique_count)]
        return ensemble_majority_dict

    def forward(self, batch_dict: Dict) -> Dict:
        ensemble_model_dict = {"output": {}}
        ensemble_pred = []
        for ensemble_idx, ensemble_net in enumerate(self.nets):
            ensemble_model_dict["output"]["ensemble_output_" + str(ensemble_idx)] = copy.deepcopy(
                ensemble_net(batch_dict)["output"]
            )
            ensemble_pred.append(ensemble_model_dict["output"]["ensemble_output_" + str(ensemble_idx)])
        ensemble_model_dict["output"].update(self.calculate_ensemble_avg(ensemble_pred))
        ensemble_model_dict["output"].update(self.calculate_ensemble_majority_vote(ensemble_pred))
        batch_dict["model"] = ensemble_model_dict
        return batch_dict



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from typing import Sequence, Dict, Tuple

import torch

from fuse.dl.models import ModelMultiHead
from fuse.utils.ndict import NDict


class ModelSiamese(ModelMultiHead):
    """
    Fuse Siamese model - 2 branches of the same convolutional neural network with multiple heads
    """

    def __init__(
        self,
        conv_inputs_0: Tuple[Tuple[str, int], ...],
        conv_inputs_1: Tuple[Tuple[str, int], ...],
        backbone: torch.nn.Module,
        heads: Sequence[torch.nn.Module] = None,
    ) -> None:
        """
        Fuse Siamese model -  two branches with same convolutional neural network with multiple heads
        :param conv_inputs_0:     batch_dict name for model input and its number of input channels
            for example: conv_inputs_0=(('data.input.input_0.tensor', 1),)
        :param conv_inputs_1:    batch_dict name for model input and its number of input channels
            for example: conv_inputs_1=(('data.input.input_1.tensor', 1),)
        :param backbone:        PyTorch backbone module - a convolutional neural network
        :param heads:           Sequence of head modules
            for example: (HeadGlobalPoolingClassifier(),)
        """
        assert heads is not None, "You must provide 'heads'"
        super().__init__((), backbone, heads)
        assert conv_inputs_0 is not None, "you must provide 'conv_inputs_0'"
        self.conv_inputs_0 = conv_inputs_0
        assert conv_inputs_1 is not None, "you must provide 'conv_inputs_1'"
        self.conv_inputs_1 = conv_inputs_1

    def forward(self, batch_dict: NDict) -> Dict:
        backbone_input_0 = [batch_dict[conv_input[0]] for conv_input in self.conv_inputs_0]  # [batch,channels(1),h,w]
        backbone_input_1 = [batch_dict[conv_input[0]] for conv_input in self.conv_inputs_1]  # [batch,channels(1),h,w]

        backbone_features_0 = self.backbone.forward(torch.stack(backbone_input_0, dim=1))  # batch, features, h', w'
        backbone_features_1 = self.backbone.forward(torch.stack(backbone_input_1, dim=1))  # batch, features, h', w'

        backbone_features = torch.cat([backbone_features_0, backbone_features_1], dim=1)

        batch_dict["model.backbone_features"] = backbone_features
        batch_dict["model.backbone_features_0"] = backbone_features_0
        batch_dict["model.backbone_features_1"] = backbone_features_1

        for head in self.heads:
            batch_dict = head.forward(batch_dict)

        return batch_dict



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import os
import tempfile
import unittest
import pandas as pds
import numpy as np
from tqdm.std import tqdm
import torchvision
from torchvision import transforms
from torch.utils.data.dataloader import DataLoader

from fuse.utils import Seed

from fuse.data.ops.ops_read import OpReadDataframe
from fuse.data.datasets.dataset_default import DatasetDefault
from fuse.data.datasets.dataset_wrap_seq_to_dict import DatasetWrapSeqToDict
from fuse.data.utils.collates import CollateDefault
from fuse.data.pipelines.pipeline_default import PipelineDefault
from fuse.data.utils.samplers import BatchSamplerDefault


class TestSamplers(unittest.TestCase):
    def setUp(self):
        pass

    def test_balanced_dataset(self):
        Seed.set_seed(1234)

        transform = transforms.Compose([transforms.ToTensor(), transforms.Normalize((0.1307,), (0.3081,))])
        # Create dataset
        mnist_data_path = os.environ.get("MNIST_DATA_PATH", tempfile.mkdtemp(prefix="mnist"))
        torch_dataset = torchvision.datasets.MNIST(mnist_data_path, download=True, train=True, transform=transform)
        print(f"torch dataset size = {len(torch_dataset)}")

        num_classes = 10
        num_samples = len(torch_dataset)

        # wrapping torch dataset
        dataset = DatasetWrapSeqToDict(name="test", dataset=torch_dataset, sample_keys=("data.image", "data.label"))
        dataset.create()
        print(dataset.summary())
        batch_sampler = BatchSamplerDefault(
            dataset=dataset,
            balanced_class_name="data.label",
            num_balanced_classes=num_classes,
            batch_size=32,
            mode="approx",
            balanced_class_weights=[1 / num_classes] * num_classes,
            workers=10,
        )

        labels = np.zeros(num_classes)

        # Create dataloader
        dataloader = DataLoader(
            dataset=dataset, collate_fn=CollateDefault(), batch_sampler=batch_sampler, shuffle=False, drop_last=False
        )
        iter1 = iter(dataloader)
        for _ in tqdm(range(len(dataloader))):
            batch_dict = next(iter1)
            labels_in_batch = batch_dict["data.label"]
            for label in labels_in_batch:
                labels[label] += 1

        # final balance
        print(labels)
        for idx in range(num_classes):
            sampled = labels[idx] / num_samples
            print(f"Class {idx}: {sampled * 100}% of data")
            self.assertAlmostEqual(
                sampled,
                1 / num_classes,
                delta=1 / num_classes * 0.5,
                msg=f"Unbalanced class {idx}, expected 0.1+-0.05 and got {sampled}",
            )

    def test_not_equalbalance_dataset(self):
        Seed.set_seed(1234)

        transform = transforms.Compose([transforms.ToTensor(), transforms.Normalize((0.1307,), (0.3081,))])
        # Create dataset
        mnist_data_path = os.environ.get("MNIST_DATA_PATH", tempfile.mkdtemp(prefix="mnist"))
        torch_dataset = torchvision.datasets.MNIST(mnist_data_path, download=True, train=True, transform=transform)
        print(f"torch dataset size = {len(torch_dataset)}")

        num_classes = 10
        probs = 1 / num_classes

        # wrapping torch dataset
        dataset = DatasetWrapSeqToDict(name="test", dataset=torch_dataset, sample_keys=("data.image", "data.label"))
        dataset.create()

        balanced_class_weights = [1] * 5 + [3] * 5
        batch_size = 20
        batch_sampler = BatchSamplerDefault(
            dataset=dataset,
            balanced_class_name="data.label",
            num_balanced_classes=num_classes,
            batch_size=batch_size,
            mode="exact",
            balanced_class_weights=balanced_class_weights,
        )

        # Create dataloader
        labels = np.zeros(num_classes)
        dataloader = DataLoader(
            dataset=dataset, collate_fn=CollateDefault(), batch_sampler=batch_sampler, shuffle=False, drop_last=False
        )
        iter1 = iter(dataloader)
        num_items = 0
        for _ in tqdm(range(len(dataloader))):
            batch_dict = next(iter1)
            labels_in_batch = batch_dict["data.label"]
            for label in labels_in_batch:
                labels[label] += 1
                num_items += 1

        # final balance
        print(labels)
        for idx in range(num_classes):
            sampled = labels[idx] / num_items
            print(f"Class {idx}: {sampled * 100}% of data")
            self.assertEqual(sampled, balanced_class_weights[idx] / batch_size)

    def test_sampler_default(self):
        # datainfo
        data = {
            "sample_id": ["a", "b", "c", "d", "e"],
            "data.values": [7, 4, 9, 2, 4],
            "data.class": [0, 1, 2, 0, 0],
        }
        df = pds.DataFrame(data)

        # create simple pipeline
        op_df = OpReadDataframe(df)
        pipeline = PipelineDefault("test", [(op_df, {})])

        # create dataset
        dataset = DatasetDefault(data["sample_id"], dynamic_pipeline=pipeline)
        dataset.create()

        # create sampler
        batch_sampler = BatchSamplerDefault(
            dataset, batch_size=3, balanced_class_name="data.class", num_balanced_classes=3, workers=0
        )

        # Use the collate function
        dl = DataLoader(dataset, collate_fn=CollateDefault(), batch_sampler=batch_sampler)
        batch = next(iter(dl))

        # verify
        self.assertEqual(len(batch_sampler), 3)
        self.assertIn(0, batch["data.class"])
        self.assertIn(1, batch["data.class"])
        self.assertIn(2, batch["data.class"])


if __name__ == "__main__":
    unittest.main()



from enum import Enum
from fuse.data.key_types import DataTypeBasic, TypeDetectorPatternsBased


class DataTypeImaging(Enum):
    """
    Possible data types stored in sample_dict.
    Using Patterns - the type will be inferred from the key name
    """

    IMAGE = "image"  # Image
    SEG = "seg"  # Segmentation Map
    BBOX = "bboxes"  # Bounding Box
    CTR = "contours"  # Contour


PATTERNS_DICT_IMAGING = {
    r".*img$": DataTypeImaging.IMAGE,
    r".*seg$": DataTypeImaging.SEG,
    r".*bbox$": DataTypeImaging.BBOX,
    r".*ctr$": DataTypeImaging.CTR,
    r".*$": DataTypeBasic.UNKNOWN,
}

type_detector_imaging = TypeDetectorPatternsBased(PATTERNS_DICT_IMAGING)



import os
from fuse.utils.remote_execution.remote_execution import (
    RemoteExecution,
    RemoteCommand,
    get_script_runner_path,
)


__all__ = [
    "RemoteExecution",
    "RemoteCommand",
    "get_script_runner_path",
]



from typing import Optional, Sequence, Tuple
from fuse.data import OpBase
from fuse.data.utils.collates import CollateDefault
from fuse.data.utils.samplers import BatchSamplerDefault
from fuse.data.utils.split import dataset_balanced_division_to_folds

from fuseimg.datasets.isic import ISIC

from torch.utils.data.dataloader import DataLoader


def isic_2019_dataloaders(
    data_path: str,
    cache_path: str,
    reset_cache: bool = False,
    reset_split_file: bool = False,
    append_dyn_pipeline: Optional[Sequence[Tuple[OpBase, dict]]] = None,
    sample_ids: Optional[Sequence[str]] = None,
) -> Tuple[DataLoader, DataLoader]:
    """
    Create train_dataloader and validation_dataloadr with specific parameters for image_clinical_multimodality tutorial
    :param data_path: path to download the data to
    :param cache_path: path to store the cached files
    :param reset_cache: restart cache or not
    :param reset_split_file: reuse previous split or create a new one.
    :param append_dyn_pipeline: steps to append at the end of the dynamic pipeline (doesn't require recaching)
    :param sample_ids: list of sample_ids tp include - otherwise will consider all the sample_ids
    :return: train_dataloader and validation_dataloader
    """
    # internal arguments used for this example
    n_folds = 3
    train_folds = [0, 1]
    validation_folds = [2]
    batch_size = 8
    num_workers = 4
    data_split_filename = "data_split.pkl"
    # split to folds randomly
    all_dataset = ISIC.dataset(
        data_path,
        cache_path,
        reset_cache=reset_cache,
        append_dyn_pipeline=append_dyn_pipeline,
        num_workers=num_workers,
        samples_ids=sample_ids,
    )

    folds = dataset_balanced_division_to_folds(
        dataset=all_dataset,
        output_split_filename=data_split_filename,
        keys_to_balance=["data.label"],
        nfolds=n_folds,
        reset_split=reset_split_file,
    )

    train_sample_ids = []
    for fold in train_folds:
        train_sample_ids += folds[fold]
    validation_sample_ids = []
    for fold in validation_folds:
        validation_sample_ids += folds[fold]

    train_dataset = ISIC.dataset(
        data_path, cache_path, samples_ids=train_sample_ids, append_dyn_pipeline=append_dyn_pipeline, train=True
    )

    print("- Create sampler:")
    sampler = BatchSamplerDefault(
        dataset=train_dataset, balanced_class_name="data.label", num_balanced_classes=8, batch_size=batch_size
    )
    print("- Create sampler: Done")

    # Create dataloader
    train_dataloader = DataLoader(
        dataset=train_dataset, batch_sampler=sampler, collate_fn=CollateDefault(), num_workers=num_workers
    )

    # dataset
    validation_dataset = ISIC.dataset(
        data_path, cache_path, samples_ids=validation_sample_ids, append_dyn_pipeline=append_dyn_pipeline, train=False
    )

    # dataloader
    validation_dataloader = DataLoader(
        dataset=validation_dataset, batch_size=batch_size, collate_fn=CollateDefault(), num_workers=num_workers
    )

    return train_dataloader, validation_dataloader


SEX_INDEX = {"male": 0, "female": 1, "N/A": 2}
ANATOM_SITE_INDEX = {
    "anterior torso": 0,
    "upper extremity": 1,
    "posterior torso": 2,
    "lower extremity": 3,
    "lateral torso": 4,
    "head/neck": 5,
    "palms/soles": 6,
    "oral/genital": 7,
    "N/A": 8,
}



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""
from functools import partial
import os
from typing import Hashable, List, Optional, Sequence, Tuple, Union

import numpy as np
import torch
from tqdm import tqdm
import skimage
import skimage.transform

from fuse.utils import NDict
from fuse.utils.rand.param_sampler import RandBool, RandInt, Uniform
import wget

from fuse.data import DatasetDefault
from fuse.data.datasets.caching.samples_cacher import SamplesCacher
from fuse.data import PipelineDefault, OpSampleAndRepeat, OpToTensor, OpRepeat
from fuse.data.ops.op_base import OpReversibleBase
from fuse.data.ops.ops_aug_common import OpSample
from fuse.data.ops.ops_common import OpLambda

from fuse.data.utils.sample import get_sample_id

from fuseimg.data.ops.aug.color import OpAugColor
from fuseimg.data.ops.aug.geometry import OpAugAffine2D
from fuseimg.data.ops.image_loader import OpLoadImage
from fuseimg.data.ops.color import OpClip, OpToRange


class OpKits21SampleIDDecode(OpReversibleBase):
    """
    decodes sample id into image and segmentation filename
    """

    def __call__(self, sample_dict: NDict, op_id: Optional[str]) -> NDict:
        """ """
        sid = get_sample_id(sample_dict)

        img_filename_key = "data.input.img_path"
        sample_dict[img_filename_key] = os.path.join(sid, "imaging.nii.gz")

        seg_filename_key = "data.gt.seg_path"
        sample_dict[seg_filename_key] = os.path.join(sid, "aggregated_MAJ_seg.nii.gz")

        return sample_dict

    def reverse(self, sample_dict: dict, key_to_reverse: str, key_to_follow: str, op_id: Optional[str]) -> dict:
        return sample_dict


def my_resize(input_tensor: torch.Tensor, resize_to: Tuple[int, int, int]) -> torch.Tensor:
    """
    Custom resize operation for the CT image
    """

    inner_image_height = input_tensor.shape[0]
    inner_image_width = input_tensor.shape[1]
    inner_image_depth = input_tensor.shape[2]
    h_ratio = resize_to[0] / inner_image_height
    w_ratio = resize_to[1] / inner_image_width
    if h_ratio >= 1 and w_ratio >= 1:
        resize_ratio_xy = min(h_ratio, w_ratio)
    elif h_ratio < 1 and w_ratio < 1:
        resize_ratio_xy = max(h_ratio, w_ratio)
    else:
        resize_ratio_xy = 1
    # resize_ratio_z = self.resize_to[2] / inner_image_depth
    if resize_ratio_xy != 1 or inner_image_depth != resize_to[2]:
        input_tensor = skimage.transform.resize(
            input_tensor,
            output_shape=(
                int(inner_image_height * resize_ratio_xy),
                int(inner_image_width * resize_ratio_xy),
                int(resize_to[2]),
            ),
            mode="reflect",
            anti_aliasing=True,
        )
    return input_tensor


class KITS21:
    """
    2021 Kidney and Kidney Tumor Segmentation Challenge Dataset
    KITS21 data pipeline implementation. See https://github.com/neheller/kits21
    Currently including only the image and segmentation map
    """

    # bump whenever the static pipeline modified
    KITS21_DATASET_VER = 0

    @staticmethod
    def download(path: str, cases: Optional[Union[int, List[int]]] = None) -> None:
        """
        :param cases: pass None (default) to download all 300 cases. OR
        pass a list of integers with cases num in the range [0,299]. OR
        pass a single int to download a single case
        """
        if cases is None:
            cases = list(range(300))
        elif isinstance(cases, int):
            cases = [cases]
        elif not isinstance(cases, list):
            raise Exception("Unsupported args! please provide None, int or list of ints")

        dl_dir = path

        for i in tqdm(cases, total=len(cases)):
            destination_dir = os.path.join(dl_dir, f"case_{i:05d}")
            os.makedirs(destination_dir, exist_ok=True)

            # imaging
            destination_file = os.path.join(destination_dir, "imaging.nii.gz")
            src = f"https://kits19.sfo2.digitaloceanspaces.com/master_{i:05d}.nii.gz"
            if not os.path.exists(destination_file):
                wget.download(src, destination_file)
            else:
                print(f"imaging.nii.gz number {i} was found")

            # segmentation
            seg_file = "aggregated_MAJ_seg.nii.gz"
            destination_file = os.path.join(destination_dir, seg_file)
            src = f"https://github.com/neheller/kits21/raw/master/kits21/data/case_{i:05d}/aggregated_MAJ_seg.nii.gz"
            if not os.path.exists(destination_file):
                wget.download(src, destination_file)
            else:
                print(f"{seg_file} number {i} was found")

    @staticmethod
    def sample_ids():
        """
        get all the sample ids in trainset
        sample_id is case_{id:05d} (for example case_00001 or case_00100)
        """
        return [f"case_{case_id:05d}" for case_id in range(300)]

    @staticmethod
    def static_pipeline(data_path: str) -> PipelineDefault:
        """
        Get suggested static pipeline (which will be cached), typically loading the data plus design choices that we won't experiment with.
        :param data_path: path to original kits21 data (can be downloaded by KITS21.download())
        """
        static_pipeline = PipelineDefault(
            "static",
            [
                # decoding sample ID
                (
                    OpKits21SampleIDDecode(),
                    dict(),
                ),  # will save image and seg path to "data.input.img_path", "data.gt.seg_path"
                # loading data
                (OpLoadImage(data_path), dict(key_in="data.input.img_path", key_out="data.input.img", format="nib")),
                (OpLoadImage(data_path), dict(key_in="data.gt.seg_path", key_out="data.gt.seg", format="nib")),
                # fixed image normalization
                (OpClip(), dict(key="data.input.img", clip=(-500, 500))),
                (OpToRange(), dict(key="data.input.img", from_range=(-500, 500), to_range=(0, 1))),
                # transposing so the depth channel will be first
                (
                    OpLambda(partial(np.moveaxis, source=-1, destination=0)),
                    dict(key="data.input.img"),
                ),  # convert image from shape [H, W, D] to shape [D, H, W]
            ],
        )
        return static_pipeline

    @staticmethod
    def dynamic_pipeline():
        """
        Get suggested dynamic pipeline. including pre-processing that might be modified and augmentation operations.
        """
        repeat_for = [dict(key="data.input.img"), dict(key="data.gt.seg")]

        dynamic_pipeline = PipelineDefault(
            "dynamic",
            [
                # resize image to (110, 256, 256)
                (
                    OpRepeat(
                        OpLambda(func=partial(my_resize, resize_to=(110, 256, 256))), kwargs_per_step_to_add=repeat_for
                    ),
                    dict(),
                ),
                # Numpy to tensor
                (OpRepeat(OpToTensor(), kwargs_per_step_to_add=repeat_for), dict(dtype=torch.float32)),
                # affine transformation per slice but with the same arguments
                (
                    OpSampleAndRepeat(OpAugAffine2D(), kwargs_per_step_to_add=repeat_for),
                    dict(
                        rotate=Uniform(-180.0, 180.0),
                        scale=Uniform(0.8, 1.2),
                        flip=(RandBool(0.5), RandBool(0.5)),
                        translate=(RandInt(-15, 15), RandInt(-15, 15)),
                    ),
                ),
                # color augmentation - check if it is useful in CT images
                (
                    OpSample(OpAugColor()),
                    dict(
                        key="data.input.img",
                        gamma=Uniform(0.8, 1.2),
                        contrast=Uniform(0.9, 1.1),
                        add=Uniform(-0.01, 0.01),
                    ),
                ),
                # add channel dimension -> [C=1, D, H, W]
                (OpLambda(partial(torch.unsqueeze, dim=0)), dict(key="data.input.img")),
            ],
        )
        return dynamic_pipeline

    @staticmethod
    def dataset(
        data_path: str,
        cache_dir: str,
        reset_cache: bool = False,
        num_workers: int = 10,
        sample_ids: Optional[Sequence[Hashable]] = None,
    ) -> DatasetDefault:
        """
        Get cached dataset
        :param data_path: path to store the original data
        :param cache_dir: path to store the cache
        :param reset_cache: set to True tp reset the cache
        :param num_workers: number of processes used for caching
        :param sample_ids: dataset including the specified sample_ids or None for all the samples. sample_id is case_{id:05d} (for example case_00001 or case_00100).
        """

        if sample_ids is None:
            sample_ids = KITS21.sample_ids()

        static_pipeline = KITS21.static_pipeline(data_path)
        dynamic_pipeline = KITS21.dynamic_pipeline()

        cacher = SamplesCacher(
            f"kits21_cache_ver{KITS21.KITS21_DATASET_VER}",
            static_pipeline,
            [cache_dir],
            restart_cache=reset_cache,
            workers=num_workers,
        )

        my_dataset = DatasetDefault(
            sample_ids=sample_ids,
            static_pipeline=static_pipeline,
            dynamic_pipeline=dynamic_pipeline,
            cacher=cacher,
        )
        my_dataset.create()
        return my_dataset



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from typing import Union

import torch
import torch.nn as nn


class Checkpoint:
    def __init__(self, net: Union[nn.Module, dict], epoch_idx: int, learning_rate: float):
        if isinstance(net, nn.Module):
            self.net_state_dict = net.state_dict()
        else:
            # we have a state dict
            self.net_state_dict = net
        self.epoch_idx = epoch_idx
        self.learning_rate = learning_rate
        pass

    def as_dict(self):
        return {"net_state_dict": self.net_state_dict, "epoch_idx": self.epoch_idx, "learning_rate": self.learning_rate}

    def save_to_file(self, file_name: str):
        torch.save(self.as_dict(), file_name)

    @classmethod
    def load_from_file(cls, file_name: str):
        checkpoint_dict = torch.load(file_name, map_location="cpu")
        net_state_dict = checkpoint_dict["net_state_dict"]
        epoch_idx = checkpoint_dict["epoch_idx"]
        learning_rate = checkpoint_dict["learning_rate"]

        return cls(net=net_state_dict, epoch_idx=epoch_idx, learning_rate=learning_rate)



from fuse.data.ops.ops_common import OpApplyTypes
from fuse.data.key_types_for_testing import type_detector_for_testing
from functools import partial

OpApplyTypesImaging = partial(
    OpApplyTypes,
    type_detector=type_detector_for_testing,
)



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

import unittest

from fuse.utils.rand.seed import Seed
from fuse.utils.file_io.file_io import save_hdf5_safe, load_hdf5
import numpy as np
import tempfile
import os


class TestHDF5IO(unittest.TestCase):
    """
    Test HDF5 IO
    """

    def setUp(self):
        pass

    def _generate_test_data_1(self):
        Seed.set_seed(1337)
        ans = {
            "data.cc.img": np.random.rand(30, 200, 200),
            "data.cc.seg": np.random.randint(0, 16, size=(30, 200, 200)),
            "data.mlo.img": np.random.rand(30, 200, 200),
            "data.mlo.seg": np.random.randint(0, 16, size=(40, 100, 164)),
            "data.clinical_info_input": np.random.rand(1000),
        }
        return ans

    def test_object_requires_hdf5_recurse(self):
        data = self._generate_test_data_1()
        tmpdir = tempfile.gettempdir()
        filename = os.path.join(tmpdir, "test_hdf5_1.hdf5")

        save_hdf5_safe(filename, **data)
        loaded_hdf5_all = load_hdf5(filename)
        loaded_hdf5_partial = load_hdf5(
            filename,
            custom_extract={
                "data.cc.img": [slice(1, 10, 2), slice(None, 20), Ellipsis],
                "data.cc.seg": None,
                "data.mlo.img": 20,
                "data.clinical_info_input": None,
            },
        )

        self.assertEqual(len(loaded_hdf5_all.keys()), 5)
        self.assertAlmostEqual(loaded_hdf5_all["data.cc.img"].sum(), 599968.1459841608)
        self.assertEqual(loaded_hdf5_all["data.cc.seg"].sum(), 9005579)
        self.assertAlmostEqual(loaded_hdf5_all["data.mlo.img"].sum(), 599511.3034792768)

        self.assertEqual(len(loaded_hdf5_partial.keys()), 4)
        self.assertAlmostEqual(loaded_hdf5_partial["data.cc.img"].sum(), 9991.398330095684)
        self.assertEqual(loaded_hdf5_partial["data.cc.seg"].sum(), 9005579)
        self.assertAlmostEqual(loaded_hdf5_partial["data.mlo.img"].sum(), 20010.11375657579)
        self.assertAlmostEqual(loaded_hdf5_partial["data.clinical_info_input"].sum(), 507.3055890687598)

    def tearDown(self):
        pass


if __name__ == "__main__":
    unittest.main()



"""
(C) Copyright 2021 IBM Corp.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
   http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
Created on June 30, 2021
"""

import os
import pathlib
from typing import Any, Dict
from collections import OrderedDict
from pycocotools.coco import COCO
import numpy as np
import nibabel as nib
from fuse.utils import set_seed

from fuse.eval.metrics.segmentation.metrics_segmentation_common import (
    MetricDice,
    MetricIouJaccard,
    MetricOverlap,
    Metric2DHausdorff,
    MetricPixelAccuracy,
)
from fuse.eval.metrics.segmentation.metrics_instance_segmentation_common import MetricInstanceIouJaccard
from fuse.eval.metrics.detection.metrics_detection_common import MetricDetectionPrecision, MetricDetectionRecall
from fuse.eval.evaluator import EvaluatorDefault


def example_seg_0() -> Dict[str, Any]:
    """
    Simple evaluation example for semantic segmentation on 3 separate binary segmentations
    Inputs are 4 pairs of segmentation files: one including predictions and one targets
    new classes mapping is defined based on the the basic pixel segmentation values
    """
    LABEL_MAPPING = {
        "label_1": (1, 2, 3),
        "label_2": (2, 3),
        "label_3": (2,),
    }

    # pre collect function to change the format
    def pre_collect_process_label_1(sample_dict: dict) -> dict:
        # convert scores to numpy array
        mask = np.zeros(sample_dict["pred.array"].shape, dtype=bool)
        mask_label = np.zeros(sample_dict["pred.array"].shape, dtype=bool)
        for l in LABEL_MAPPING["label_1"]:
            mask[sample_dict["pred.array"] == l] = True
            mask_label[sample_dict["label.array"] == l] = True
        sample_dict["pred.array_label_1"] = mask
        sample_dict["label.array_label_1"] = mask_label
        return sample_dict

    def pre_collect_process_label_3(sample_dict: dict) -> dict:
        # convert scores to numpy array
        mask_pred = np.zeros(sample_dict["pred.array"].shape, dtype=bool)
        mask_label = np.zeros(sample_dict["pred.array"].shape, dtype=bool)
        for l in LABEL_MAPPING["label_3"]:
            mask_pred[sample_dict["pred.array"] == l] = True
            mask_label[sample_dict["label.array"] == l] = True
        sample_dict["pred.array_label_3"] = mask_pred
        sample_dict["label.array_label_3"] = mask_label
        return sample_dict

    def pre_collect_process_label_2(sample_dict: dict) -> dict:
        # convert scores to numpy array
        mask = np.zeros(sample_dict["pred.array"].shape, dtype=bool)
        mask_label = np.zeros(sample_dict["pred.array"].shape, dtype=bool)
        for l in LABEL_MAPPING["label_2"]:
            mask[sample_dict["pred.array"] == l] = True
            mask_label[sample_dict["label.array"] == l] = True
        sample_dict["pred.array_label_2"] = mask
        sample_dict["label.array_label_2"] = mask_label
        return sample_dict

    # define iterator
    def data_iter():
        dir_path = pathlib.Path(__file__).parent.resolve()
        predicted_list = os.listdir(os.path.join(dir_path, "inputs/semantic_segmentation/predicted/"))
        labels_path = os.path.join(dir_path, "inputs/semantic_segmentation/labeled/")
        for predicted in predicted_list:
            id = os.path.basename(predicted).split(".")[0]
            label_path = os.path.join(labels_path, id, "seg.nii.gz")
            predicted_path = os.path.join(dir_path, "inputs/semantic_segmentation/predicted/", predicted)
            sample_dict = {}
            sample_dict["id"] = id
            sample_dict["pred.array"] = np.asanyarray(nib.load(predicted_path).dataobj)
            sample_dict["label.array"] = np.asanyarray(nib.load(label_path).dataobj)
            yield sample_dict

    # list of metrics
    metrics = OrderedDict(
        [
            (
                "dice_label_1",
                MetricDice(
                    pred="pred.array_label_1",
                    target="label.array_label_1",
                    pre_collect_process_func=pre_collect_process_label_1,
                ),
            ),
            (
                "dice_label_2",
                MetricDice(
                    pred="pred.array_label_2",
                    target="label.array_label_2",
                    pre_collect_process_func=pre_collect_process_label_2,
                ),
            ),
            (
                "dice_label_3",
                MetricDice(
                    pred="pred.array_label_3",
                    target="label.array_label_3",
                    pre_collect_process_func=pre_collect_process_label_3,
                ),
            ),
        ]
    )

    evaluator = EvaluatorDefault()
    results = evaluator.eval(ids=None, data=data_iter(), batch_size=1, metrics=metrics)

    return results


def example_seg_1() -> Dict[str, Any]:
    """
    Simple evaluation example for binary segmentation on metric score such as dice,overlap,pixel_accuracy,iou_jacard
    Inputs are pair of pixel array : one including predictions and one targets
    """

    # define iterator
    def data_iter():
        # set seed
        set_seed(0)

        sample_dict = {}
        sample_dict["id"] = id
        sample_dict["pred.array"] = np.random.rand(10, 5, 5, 4) > 0.5
        sample_dict["label.array"] = np.random.rand(10, 5, 5, 4) > 0.5
        yield sample_dict

    # list of metrics
    metrics = OrderedDict(
        [
            ("dice", MetricDice(pred="pred.array", target="label.array")),
            ("overlap", MetricOverlap(pred="pred.array", target="label.array")),
            ("pixel_accuracy", MetricPixelAccuracy(pred="pred.array", target="label.array")),
            ("iou_jaccard", MetricIouJaccard(pred="pred.array", target="label.array")),
        ]
    )

    evaluator = EvaluatorDefault()
    results = evaluator.eval(ids=None, data=data_iter(), batch_size=60, metrics=metrics)

    return results


def example_seg_2() -> Dict[str, Any]:
    """
    Simple evaluation example for dice score for multiclass semantic segmentation
    Inputs are 4 pairs of segmentation files: one including predictions and one targets
    """
    # define iterator
    def data_iter():
        dir_path = pathlib.Path(__file__).parent.resolve()
        predicted_list = os.listdir(os.path.join(dir_path, "inputs/semantic_segmentation/predicted/"))
        labels_path = os.path.join(dir_path, "inputs/semantic_segmentation/labeled/")
        for predicted in predicted_list:
            id = os.path.basename(predicted).split(".")[0]
            label_path = os.path.join(labels_path, id, "seg.nii.gz")
            predicted_path = os.path.join(dir_path, "inputs/semantic_segmentation/predicted/", predicted)
            sample_dict = {}
            sample_dict["id"] = id
            sample_dict["pred.array"] = np.asanyarray(nib.load(predicted_path).dataobj)
            sample_dict["label.array"] = np.asanyarray(nib.load(label_path).dataobj)
            yield sample_dict

    # list of metrics
    metrics = OrderedDict(
        [
            ("dice", MetricDice(pred="pred.array", target="label.array")),
        ]
    )

    evaluator = EvaluatorDefault()
    results = evaluator.eval(ids=None, data=data_iter(), batch_size=1, metrics=metrics)

    return results


def example_seg_3() -> Dict[str, Any]:
    """
    Simple evaluation example for 2d hausdorff distance for segmentation and the support of pixel weight in dice,overlap,pixel_accuracu,iou_jaccard
    Inputs are pair of pixel arrays : one including predictions and one targets
    """

    # define iterator
    def data_iter():
        sample_dict = {}
        sample_dict["id"] = id
        sample_dict["pred.array"] = np.array([(1.0, 0.0), (0.0, 1.0), (1.0, 0.0), (0.0, 1.0)])
        sample_dict["label.array"] = np.array([(1.0, 1.0), (1.0, 1.0), (1.0, 1.0), (1.0, 1.0)])
        sample_dict["pixel_weight"] = {"1": np.array([(0.125, 0.125), (0.125, 0.125), (0.125, 0.125), (0.125, 0.125)])}
        yield sample_dict

    # list of metrics
    metrics = OrderedDict(
        [
            ("hausdorff", Metric2DHausdorff(pred="pred.array", target="label.array", class_weights={"1": 1.0})),
            (
                "dice",
                MetricDice(
                    pred="pred.array", target="label.array", pixel_weight="pixel_weight", class_weights={"1": 1.0}
                ),
            ),
            (
                "overlap",
                MetricOverlap(
                    pred="pred.array", target="label.array", pixel_weight="pixel_weight", class_weights={"1": 1.0}
                ),
            ),
            (
                "pixel_accuracy",
                MetricPixelAccuracy(
                    pred="pred.array", target="label.array", pixel_weight="pixel_weight", class_weights={"1": 1.0}
                ),
            ),
            (
                "iou_jaccard",
                MetricIouJaccard(
                    pred="pred.array", target="label.array", pixel_weight="pixel_weight", class_weights={"1": 1.0}
                ),
            ),
        ]
    )

    evaluator = EvaluatorDefault()
    results = evaluator.eval(ids=None, data=data_iter(), batch_size=60, metrics=metrics)

    return results


def example_seg_4() -> Dict[str, Any]:
    """
    Simple evaluation example for instance segmentation and detection based on COCO dataset
    Inputs are pair of pixel arrays : one including predictions and one targets
    """

    # define iterator
    def data_iter():
        dir_path = pathlib.Path(__file__).parent.resolve()
        annotation_path = os.path.join(dir_path, "inputs/detection/example_coco_new.json")
        cocoGt = COCO(annotation_path)
        # initialize COCO detections api
        resFile = os.path.join(dir_path, "inputs/detection/instances_val2014_fakesegm100_results.json")
        coco = cocoGt.loadRes(resFile)
        catNms = ["person", "car"]
        segtypes = ["bbox", "polygon"]
        map_field = {"polygon": "segmentation", "bbox": "bbox"}
        catIds = coco.getCatIds(catNms)
        imgIds = coco.getImgIds(catIds=catIds)
        for img_id in imgIds:
            img = coco.loadImgs(ids=[img_id])[0]
            sample_dict = {}
            sample_dict["id"] = id
            sample_dict["height"] = img["height"]
            sample_dict["width"] = img["width"]
            for index, catID in enumerate(catIds):
                pred_annIds = coco.getAnnIds(imgIds=img_id, catIds=[catID], iscrowd=None)
                target_annIds = cocoGt.getAnnIds(imgIds=img_id, catIds=[str(catID)], iscrowd=None)
                for type in segtypes:
                    pred_annotations = [seg[map_field[type]] for seg in coco.loadAnns(pred_annIds)]
                    target_annotations = [seg[map_field[type]] for seg in cocoGt.loadAnns(target_annIds)]
                    sample_dict[f"pred.array_{type}_{catNms[index]}"] = pred_annotations
                    sample_dict[f"label.array_{type}_{catNms[index]}"] = target_annotations
            yield sample_dict

    # list of metrics
    metrics = OrderedDict(
        [
            (
                "iou_bbox_person",
                MetricInstanceIouJaccard(
                    pred="pred.array_bbox_person",
                    target="label.array_bbox_person",
                    segmentation_pred_type="bbox",
                    segmentation_target_type="bbox",
                    height="height",
                    width="width",
                ),
            ),
            (
                "iou_bbox_car",
                MetricInstanceIouJaccard(
                    pred="pred.array_bbox_car",
                    target="label.array_bbox_car",
                    segmentation_pred_type="bbox",
                    segmentation_target_type="bbox",
                    height="height",
                    width="width",
                ),
            ),
            (
                "recall_bbox_person",
                MetricDetectionRecall(
                    pred="pred.array_bbox_person",
                    target="label.array_bbox_person",
                    segmentation_pred_type="bbox",
                    segmentation_target_type="bbox",
                    height="height",
                    width="width",
                ),
            ),
            (
                "recall_bbox_car",
                MetricDetectionRecall(
                    pred="pred.array_bbox_car",
                    target="label.array_bbox_car",
                    segmentation_pred_type="bbox",
                    segmentation_target_type="bbox",
                    height="height",
                    width="width",
                ),
            ),
            (
                "precision_bbox_person",
                MetricDetectionPrecision(
                    pred="pred.array_bbox_person",
                    target="label.array_bbox_person",
                    segmentation_pred_type="bbox",
                    segmentation_target_type="bbox",
                    height="height",
                    width="width",
                ),
            ),
            (
                "precision_bbox_car",
                MetricDetectionPrecision(
                    pred="pred.array_bbox_car",
                    target="label.array_bbox_car",
                    segmentation_pred_type="bbox",
                    segmentation_target_type="bbox",
                    height="height",
                    width="width",
                ),
            ),
            (
                "iou_polygon_person",
                MetricInstanceIouJaccard(
                    pred="pred.array_polygon_person",
                    target="label.array_polygon_person",
                    segmentation_pred_type="compressed_RLE",
                    segmentation_target_type="polygon",
                    height="height",
                    width="width",
                ),
            ),
            (
                "iou_polygon_car",
                MetricInstanceIouJaccard(
                    pred="pred.array_polygon_car",
                    target="label.array_polygon_car",
                    segmentation_pred_type="compressed_RLE",
                    segmentation_target_type="polygon",
                    height="height",
                    width="width",
                ),
            ),
            (
                "recall_polygon_person",
                MetricDetectionRecall(
                    pred="pred.array_polygon_person",
                    target="label.array_polygon_person",
                    segmentation_pred_type="compressed_RLE",
                    segmentation_target_type="polygon",
                    height="height",
                    width="width",
                ),
            ),
            (
                "recall_polygon_car",
                MetricDetectionRecall(
                    pred="pred.array_polygon_car",
                    target="label.array_polygon_car",
                    segmentation_pred_type="compressed_RLE",
                    segmentation_target_type="polygon",
                    height="height",
                    width="width",
                ),
            ),
            (
                "precision_polygon_person",
                MetricDetectionPrecision(
                    pred="pred.array_polygon_person",
                    target="label.array_polygon_person",
                    segmentation_pred_type="compressed_RLE",
                    segmentation_target_type="polygon",
                    height="height",
                    width="width",
                ),
            ),
            (
                "precision_polygon_car",
                MetricDetectionPrecision(
                    pred="pred.array_polygon_car",
                    target="label.array_polygon_car",
                    segmentation_pred_type="compressed_RLE",
                    segmentation_target_type="polygon",
                    height="height",
                    width="width",
                ),
            ),
        ]
    )

    evaluator = EvaluatorDefault()
    results = evaluator.eval(ids=None, data=data_iter(), batch_size=60, metrics=metrics)

    return results



from typing import Any, Dict, Hashable, Optional, Sequence
from collections import Counter
from fuse.eval.metrics.metrics_common import MetricWithCollectorBase


class MetricUniqueValues(MetricWithCollectorBase):
    """
    Collect the all the categorical values and the number of occurrences
    Result format: list of tuples - each tuple include the value and number of occurrences
    """

    def __init__(self, key: str, **kwargs) -> None:
        super().__init__(key=key, **kwargs)

    def eval(self, results: Dict[str, Any] = None, ids: Optional[Sequence[Hashable]] = None) -> None:
        values = self._extract_arguments(results, ids)["key"]
        counter = Counter(values)

        return list(counter.items())



from re import T
from typing import Sequence, Union, Dict, Optional
import numpy as np
import matplotlib.pyplot as plt
import math
import torch
from torch import nn, optim
from scipy import stats


class Calibration:
    """
    Methods for probability calibration
    """

    @staticmethod
    def reliability_diagram(
        pred: Sequence[Union[np.ndarray, float]],
        target: Sequence[Union[np.ndarray, int]],
        num_bins: Optional[int] = 10,
        num_quantiles: Optional[int] = None,
        output_filename: Optional[str] = None,
    ) -> Dict:
        """
        Reliability diagram
        :param pred: List of arrays of shape [NUM_CLASSES]
        :param target: List of integers [0, NUM_CLASSES) denoting ground truth classes
        :num_bins: Integer denoting the number of equal width bins in the diagram's x-axis.
        :num_quantiles: Integer denoting the number of equal number of samples quantiles in the diagram's x-axis.
            Note, either num_bins or num_quantiles should be set to None, as they define different ways to split the x-axis.
        :output_filename: Path to file to which to save the plot
        :return: saving roc curve to a file and return the input to figure to a dictionary
        """
        assert num_bins is None or num_quantiles is None

        if isinstance(pred[0], float):  # binary case
            pred = [np.array([1 - p, p]) for p in pred]

        max_pred = [np.max(p) for p in pred]
        cls_pred = [np.argmax(p) for p in pred]

        assert all(
            [math.isclose(p.sum(), 1, rel_tol=1e-5) for p in pred]
        )  # make sure that the inputs make proper probabilities

        total_samples = len(pred)
        max_pred = np.array(max_pred)
        if num_bins is not None:
            conf_vec = np.linspace(0, 1, num_bins + 1)
        else:
            quantiles_vec = np.linspace(0, 1, num_quantiles + 1)
            conf_vec = stats.mstats.mquantiles(max_pred, quantiles_vec.tolist())
        acc_vec = np.zeros(len(conf_vec) - 1)
        fraction_of_samples = np.zeros(len(conf_vec) - 1)
        num_samples = np.zeros(len(conf_vec) - 1)
        for i in range(len(conf_vec) - 1):
            cnt_correct_in_conf = sum(
                [
                    p_cls == t
                    for (p_cls, p, t) in zip(cls_pred, max_pred, target)
                    if p >= conf_vec[i] and p < conf_vec[i + 1]
                ]
            )
            cnt_conf = len([p for p in max_pred if p >= conf_vec[i] and p < conf_vec[i + 1]])
            num_samples[i] = cnt_conf
            fraction_of_samples[i] = cnt_conf / np.max([total_samples, np.finfo(float).eps])
            acc_vec[i] = cnt_correct_in_conf / np.max([cnt_conf, np.finfo(float).eps])
        mid_conf_vec = np.diff(conf_vec) / 2 + conf_vec[:-1]
        cnt_correct_total = sum([p_cls == t for (p_cls, t) in zip(cls_pred, target)])
        avg_acc = cnt_correct_total / total_samples
        avg_conf = np.mean(max_pred)
        # extract info for the plot
        results = {}
        results["conf_vec"] = conf_vec
        results["mid_confidence_points"] = mid_conf_vec
        results["accuracy"] = acc_vec
        results["avg_accuracy"] = avg_acc
        results["avg_confidence"] = avg_conf
        results["num_samples_per_bin"] = num_samples
        results["total_samples"] = total_samples

        # display
        if output_filename is not None:
            f, (ax1, ax2) = plt.subplots(2, 1, sharex=True, sharey=True)
            f.tight_layout(pad=3.0)
            ax1.plot(mid_conf_vec, fraction_of_samples, color="blue", marker="o", linewidth=2)
            ax1.set_title(
                f"Confidence Histogram. Avg. Accuracy: {np.round(avg_acc, 2)}, Avg. Confidence: {np.round(avg_conf, 2)}"
            )
            ax1.set_xlabel("Confidence")
            ax1.set_ylabel("Fraction of samples")
            ax1.grid()
            ax2.plot(conf_vec, conf_vec, color="black", linewidth=2)
            ax2.plot(mid_conf_vec, acc_vec, color="red", marker="o", linewidth=2)
            ax2.set_title("Reliability Curve")
            ax2.set_xlabel("Confidence")
            ax2.set_ylabel("Accuracy")
            ax2.grid()
            f.savefig(output_filename)
            plt.close()

        return results

    @staticmethod
    def ece(
        pred: Sequence[Union[np.ndarray, float]],
        target: Sequence[Union[np.ndarray, int]],
        num_bins: Optional[int] = 10,
        num_quantiles: Optional[int] = None,
    ) -> float:
        """
        :param pred: prediction array per sample. Each element shape [num_classes]
        :param target: target per sample. Each element is an integer in range [0 - num_classes)
        :num_bins: Integer denoting the number of equal width bins in the diagram's x-axis.
        :num_quantiles: Integer denoting the number of equal number of samples quantiles in the diagram's x-axis.
            Note, either num_bins or num_quantiles should be set to None, as they define different ways to split the x-axis.
        :return Expected Calibration Error (ECE) score
        """
        reliability_results = Calibration.reliability_diagram(
            pred, target, num_bins=num_bins, num_quantiles=num_quantiles, output_filename=None
        )
        conf_vec = reliability_results["conf_vec"][1:]
        total_samples = reliability_results["total_samples"]
        num_samples_per_bin = reliability_results["num_samples_per_bin"]
        existing_bins = num_samples_per_bin > 0
        accuracy_vec = reliability_results["accuracy"]
        # expected calibration error
        ece = (1.0 / total_samples) * np.sum(num_samples_per_bin * np.abs(accuracy_vec - conf_vec))
        # maximum calibration error
        # we filter out bins which don't contain any samples
        mce = np.max(np.abs(accuracy_vec[existing_bins] - conf_vec[existing_bins]))

        results = {}
        results["ece"] = ece
        results["mce"] = mce

        return results

    @staticmethod
    def find_temperature(pred: Sequence[Union[np.ndarray, float]], target: Sequence[Union[np.ndarray, int]]) -> float:
        """
        :param pred: prediction array per sample. Each element shape [num_classes]
            note that pred should be logits *before* conversion to softmax probabilities
        :param target: target per sample. Each element is an integer in range [0 - num_classes)
        :return temperature scale for calibration
        We optimize to find the temperature similarly to: https://github.com/gpleiss/temperature_scaling
        """
        nll_criterion = nn.CrossEntropyLoss()
        temperature = nn.Parameter(torch.ones(1) * 1.5)
        optimizer = optim.LBFGS([temperature], lr=0.01, max_iter=50)
        with torch.no_grad():
            logits = torch.tensor(np.array(pred))
            target = torch.tensor(target)

        def eval():
            optimizer.zero_grad()
            loss = nll_criterion(logits / temperature, target)
            loss.backward()
            return loss

        optimizer.step(eval)

        return temperature.item()

    @staticmethod
    def apply_temperature(pred: Sequence[Union[np.ndarray, float]], temperature: Union[float, None] = None) -> float:
        """
        :param pred: prediction array per sample. Each element shape [num_classes]
            note that pred should be logits *before* conversion to softmax probabilities
        :param temperature: temperature scalar
        :return array of softmax probabilities
        """
        if temperature is None:
            pred_calibrated = pred
        else:
            if isinstance(pred[0], np.ndarray):
                pred = np.array(pred)
            pred_calibrated = torch.tensor(pred / temperature)
            pred_calibrated = nn.Softmax(dim=1)(pred_calibrated)

        return np.array(pred_calibrated)



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""
from typing import Dict, Optional, Sequence, Tuple, Union

import numpy as np
import scipy
from statsmodels.stats.contingency_tables import mcnemar


class ModelComparison:
    """
    Methods for model comparison
    """

    @staticmethod
    def bootstrap_margin_superiority(
        name: Optional[str],
        test_values: Sequence[float],
        reference_values: Sequence[float],
        margin: float = 0.0,
        margin_type: str = "abs",
    ) -> Dict[str, float]:
        """
        compare method - to be used in PairedBootstrap.
        Compute p_value for the hypothesis that a tested model is superior to a reference model by a margin of error.
        :param name: the original metric name - not used
        :param test_values: test values to compare to reference values. Each element represent a bootstrap iteration.
        :param reference_values: reference values. Each element represent a bootstrap iteration.
        :param margin: the difference required to consider test result superior to a reference result
        :param margin_type: "abs" - absolute margin, "rel" - relative margin (actual margin will be margin * reference_result)
        :return: results dictionary. format: {"diff": <>, "count": <>, "count_superior": <>, "p_value": <>}
        """
        test_values = np.array(test_values)
        reference_values = np.array(reference_values)

        results = {"count": len(test_values)}
        margin_abs = margin
        if margin_type == "rel":
            margin_abs *= reference_values
        diff = test_values - reference_values
        results["count_superior"] = np.where(diff > margin_abs, 1, 0).sum()
        results["diff"] = diff.sum() / results["count"]
        results["p_value"] = 1 - results["count_superior"] / results["count"]

        return results

    @staticmethod
    def bootstrap_margin_non_inferiority(
        name: Optional[str],
        test_values: Sequence[float],
        reference_values: Sequence[float],
        margin: float = 0.0,
        margin_type: str = "abs",
    ) -> Dict[str, float]:
        """
        compare method - to be used in PairedBootstrap.
        Compute p_value for the hypothesis that a tested model is not inferior to a reference model by a margin of error.
        :param name: the original metric name - not used
        :param test_values: test values to compare to reference values. Each element represent a bootstrap iteration.
        :param reference_values: reference values. Each element represent a bootstrap iteration.
        :param margin: the difference required to consider test result non inferior to a reference result
        :param margin_type: "abs" - absolute margin, "rel" - relative margin (actual margin will be margin * reference_result)
        :return: results dictionary. format: {"diff": <>, "count": <>, "count_non_inferior": <>, "p_value": <>}
        """
        test_values = np.array(test_values)
        reference_values = np.array(reference_values)

        results = {"count": len(test_values)}
        margin_abs = margin
        if margin_type == "rel":
            margin_abs *= reference_values
        diff = test_values - reference_values
        results["count_non_inferior"] = np.where(diff > -margin_abs, 1, 0).sum()
        results["diff"] = diff.sum() / results["count"]
        results["p_value"] = 1 - results["count_non_inferior"] / results["count"]

        return results

    @staticmethod
    def bootstrap_margin_equality(
        name: Optional[str],
        test_values: Sequence[float],
        reference_values: Sequence[float],
        margin: float = 0.0,
        margin_type: str = "abs",
    ) -> Dict[str, float]:
        """
        compare method - to be used in PairedBootstrap.
        Compute p_value for the hypothesis that a tested model is equal to a reference model by a margin of error.
        :param name: the original metric name - not used
        :param test_values: test values to compare to reference values. Each element represent a bootstrap iteration.
        :param reference_values: reference values. Each element represent a bootstrap iteration.
        :param margin: the difference required to consider test result equal by margin to a reference result
        :param margin_type: "abs" - absolute margin, "rel" - relative margin (actual margin will be margin * reference_result)
        :return: results dictionary. format: {"diff": <>, "count": <>, "count_margin_equal": <>, "p_value": <>}
        """
        test_values = np.array(test_values)
        reference_values = np.array(reference_values)

        results = {"count": len(test_values)}
        margin_abs = margin
        if margin_type == "rel":
            margin_abs *= reference_values
        diff = abs(test_values - reference_values)
        results["count_margin_equal"] = np.where(diff < margin_abs, 1, 0).sum()
        results["diff"] = diff.sum() / results["count"]
        results["p_value"] = 1 - results["count_margin_equal"] / results["count"]

        return results

    @staticmethod
    def delong_auc_test(
        pred1: Sequence[np.ndarray],
        pred2: Sequence[np.ndarray],
        target: Sequence[np.ndarray],
        pos_class_index: int = -1,
    ) -> Dict:
        """
        Compute p-value resulting from DeLong's statistical test to compare two binary classifiers' ROC AUCs.
        The p-value represents likelihood for the (null) hypothesis that the two ROC AUCs are similar.
        A small p-value means the models are likely to be different.
        :param pred1: list of prediction arrays per sample for the first classifier. Each element shape is [num_classes].
            for a binary classifier, can also be a list of prediction values.
        :param pred2: list of prediction arrays per sample for the second classifier. Each element shape is [num_classes].
            for a binary classifier, can also be a list of prediction values.
        :param target: target per sample. Each element is an integer in range [0 - num_classes)
        :param pos_class_index: index of the positive class (for one vs. rest).
            If the classifier is binary (no use for one vs. rest), then this parameter should be left as its default (-1).
            Otherwise, it denotes which class index is treated as positive (for one vs. rest)
        :return {'p-value', 'z', 'auc1', 'auc2', 'cov11', 'cov12', 'cov22'},
             z is the Z-score (standard score), the normal distribution value for the relevant random variable
                which in our case depends on the models' AUCs and their variance/covariance (DeLong et al. 1988)
             auc1 is the 1st model's AUC
             auc2 is the 2nd model's AUC
             cov11 is the 11 element of DeLong's covariance matrix (variance of the 1st model's AUC)
             cov12 is the 12 element of DeLong's covariance matrix (equal to the 21 element - symmetric matrix)
             cov22 is the 22 element of DeLong's covariance matrix (variance of the 2nd model's AUC)
        """
        if isinstance(pred1[0], float):
            predictions_1 = np.array(pred1)
            predictions_2 = np.array(pred2)
            ground_truth = np.array(target)
        else:
            if pos_class_index < 0:
                pos_class_index = pred1[0].shape[0] - 1
            predictions_1 = np.asarray(pred1)[:, pos_class_index]
            predictions_2 = np.asarray(pred2)[:, pos_class_index]
            ground_truth = (np.array(target) == pos_class_index) * 1

        positive_labels = ground_truth == 1
        negative_labels = ground_truth == 0
        m = (positive_labels).sum()  # number of positives
        n = (negative_labels).sum()  # number of negatives
        x1 = predictions_1[positive_labels]
        y1 = predictions_1[negative_labels]
        x2 = predictions_2[positive_labels]
        y2 = predictions_2[negative_labels]
        # convert to matrix:
        X1 = x1[:, np.newaxis].repeat(n, 1)
        X2 = x2[:, np.newaxis].repeat(n, 1)
        Y1 = y1[np.newaxis, :].repeat(m, 0)
        Y2 = y2[np.newaxis, :].repeat(m, 0)

        PSI1 = (Y1 < X1) * 1.0
        PSI1[Y1 == X1] = 0.5

        PSI2 = (Y2 < X2) * 1.0
        PSI2[Y2 == X2] = 0.5

        # empirical AUC:
        emp_auc_1 = (1.0 / (m * n)) * PSI1.sum()
        emp_auc_2 = (1.0 / (m * n)) * PSI2.sum()

        # structural components (DeLong et al. 1988):
        V10_1 = (1.0 / n) * PSI1.sum(1)
        V10_2 = (1.0 / n) * PSI2.sum(1)

        V01_1 = (1.0 / m) * PSI1.sum(0)
        V01_2 = (1.0 / m) * PSI2.sum(0)

        # covariance matrix
        S10 = (1.0 / (m - 1)) * np.array(
            [
                [np.linalg.norm(V10_1 - emp_auc_1) ** 2, (V10_1 - emp_auc_1).dot(V10_2 - emp_auc_2)],
                [(V10_2 - emp_auc_2).dot(V10_1 - emp_auc_1), np.linalg.norm(V10_2 - emp_auc_2) ** 2],
            ]
        )

        S01 = (1.0 / (n - 1)) * np.array(
            [
                [np.linalg.norm(V01_1 - emp_auc_1) ** 2, (V01_1 - emp_auc_1).dot(V01_2 - emp_auc_2)],
                [(V01_2 - emp_auc_2).dot(V01_1 - emp_auc_1), np.linalg.norm(V01_2 - emp_auc_2) ** 2],
            ]
        )

        S = (1.0 / m) * S10 + (1.0 / n) * S01

        # z-score:
        z = (emp_auc_1 - emp_auc_2) / ((S[0, 0] + S[1, 1] - 2 * S[0, 1]) ** 0.5 + np.finfo(float).eps)

        # p-value:
        p_value = 2 * scipy.stats.norm.sf(abs(z), loc=0, scale=1)

        results = {
            "p_value": p_value,
            "z": z,
            "auc1": emp_auc_1,
            "auc2": emp_auc_2,
            "cov11": S[0, 0],
            "cov12": S[0, 1],
            "cov22": S[1, 1],
        }

        return results

    @staticmethod
    def contingency_table(var1: Sequence[bool], var2: Sequence[bool]) -> np.ndarray:
        """
        Compute contingency table from two paired variables
        :param var1: first variable [list of 0's and 1's or booleans]
        :param var2: second variable [list of 0's and 1's or booleans]
        :return contingency table [2x2 numpy array]
        """
        var1 = np.array(var1, dtype=bool)
        var2 = np.array(var2, dtype=bool)

        con_tab = np.zeros((2, 2))
        con_tab[0, 0] = (np.logical_and(var1, var2)).sum()
        con_tab[0, 1] = (np.logical_and(var1, ~var2)).sum()
        con_tab[1, 0] = (np.logical_and(~var1, var2)).sum()
        con_tab[1, 1] = (np.logical_and(~var1, ~var2)).sum()

        return con_tab

    @staticmethod
    def mcnemars_test(
        pred1: Union[Sequence, np.ndarray],
        pred2: Union[Sequence, np.ndarray],
        target: Optional[Union[Sequence, np.ndarray]],
        exact: bool = True,
    ) -> Dict:
        """
        Compute p-value resulting from McNemar's statistical test to compare two models' predictions or accuracies.
        The p-value represents likelihood for the (null) hypothesis that the two paired variables are similar
        in the sense that they have a similar proportion of disagreements.
        A small p-value means they are likely to be different.
        :param pred1: predictions of the first model.
        :param pred2: predictions of the second model.
        :param target: ground truth vector
        :param exact: If True, then the binomial distribution will be used. Otherwise, the chi-square distribution, which is the approximation to the distribution of the test statistic for large sample sizes.
            The exact test is recommended for small (<25) number of discordants in the contingency table
        :return {'p-value', 'statistic', 'n1', 'n2'},
             'statistic' is the min(n1, n2), where n1, n2 are cases that are zero
                in one classifier but one in the other. This statistic is used in the exact binomial distribution test.
             'n1' and 'n2' as just defined
        """
        pred1 = np.array(pred1)
        pred2 = np.array(pred2)
        target = np.array(target)

        if target is not None:
            pred1_correct = pred1 == target
            pred2_correct = pred2 == target
            contingency_table = ModelComparison.contingency_table(pred1_correct, pred2_correct)
        else:
            contingency_table = ModelComparison.contingency_table(pred1, pred2)

        # calculate mcnemar test
        res = mcnemar(contingency_table, exact=exact)

        results = {
            "p_value": res.pvalue,
            "statistic": res.statistic,
            "n1": contingency_table[0, 1],
            "n2": contingency_table[1, 0],
        }

        return results


def psi(x, y):  # Heaviside function
    if y < x:
        return 1
    elif y == x:
        return 0.5
    else:
        return 0



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import logging
import os
import subprocess
import traceback
from typing import Any, List, Optional

import torch

from fuse.utils.utils_debug import FuseDebug


def get_available_gpu_ids() -> List[int]:
    nvidia_smi_output = run_nvidia_smi()
    if nvidia_smi_output is None:
        return None

    available_gpu_ids = get_available_gpu_ids_from_nvidia_smi_output(nvidia_smi_output)
    return available_gpu_ids


def set_cuda_visible_devices(list_of_gpu_ids: List[int]) -> None:
    devices = [str(x) for x in list_of_gpu_ids]
    allow_visible_gpus_str = ",".join(devices)
    os.environ["CUDA_VISIBLE_DEVICES"] = allow_visible_gpus_str
    print(f"os.environ['CUDA_VISIBLE_DEVICES'] = {allow_visible_gpus_str}")


def choose_and_enable_multiple_gpus(
    num_gpus_needed: int, force_gpus: Optional[List[int]] = None, use_cpu_if_fail: bool = False
) -> int:
    """
    Look for free gpus and set CUDA_VISIBLE_DEVICES accordingly
    :param num_gpus_needed: required number of gpus
    :param force_gpus: Optional, use this list instead of looking for free ones,
    :return: number of gpus allocated
    """
    # debug - num gpus
    try:
        override_num_gpus = FuseDebug().get_setting("manager_override_num_gpus")
        if override_num_gpus != "default":
            num_gpus_needed = min(override_num_gpus, num_gpus_needed)
            logging.getLogger("Fuse").info(
                f"Manager - debug mode - override num_gpus to {num_gpus_needed}", {"color": "red"}
            )

        if num_gpus_needed == 0:
            return

        if force_gpus is None:
            available_gpu_ids = get_available_gpu_ids()
        else:
            available_gpu_ids = force_gpus

        if available_gpu_ids is None:
            raise Exception("could not auto-detect available GPUs")
        elif len(available_gpu_ids) < num_gpus_needed:
            raise Exception(
                "not enough GPUs available, requested %d GPUs but only IDs %s are available!"
                % (num_gpus_needed, str(available_gpu_ids))
            )
        else:
            selected_gpu_ids = sorted(available_gpu_ids, reverse=True)[:num_gpus_needed]
            logging.getLogger("Fuse").info("selecting GPUs %s" % str(selected_gpu_ids))
            set_cuda_visible_devices(selected_gpu_ids)

        torch.backends.cudnn.benchmark = False  # to prevent gpu illegal instruction exceptions
        torch.multiprocessing.set_sharing_strategy("file_system")  # to prevent a bug of too many open file descriptors
    except Exception as e:
        if use_cpu_if_fail:
            lgr = logging.getLogger("Fuse")
            track = traceback.format_exc()
            lgr.warning(e)
            lgr.warning(track)
            return 0
        else:
            raise

    return num_gpus_needed


def move_tensor_to_device(tensor: Any, device: str) -> Any:
    """
    Moves tensor to device (if the type of tensor is torch.Tensor).
    Returns the tensor back.
    :param tensor: data to move to device
    :param device: the device
    :return: tensor
    """
    if isinstance(tensor, torch.Tensor):
        tensor = tensor.to(device)
    return tensor


def allocate_gpu_for_process(gpu_list: list):
    """
    Allocate a gpu for a process given list of available gpus shared between processes
    :param gpu_list: available gpus shared between processes
    :return: None
    """
    try:
        # extract gpu_id dedicated to that process only
        gpu_id = gpu_list.pop()
        if gpu_id == "cpu":
            gpu_id = ""
        os.environ["CUDA_VISIBLE_DEVICES"] = str(gpu_id)
    except:
        lgr = logging.getLogger("Fuse")
        # don't throw error in init function - simply print the error
        track = traceback.format_exc()
        lgr.error(track)


def deallocate_gpu():
    import gc

    with torch.no_grad():
        torch.cuda.empty_cache()
    for obj in gc.get_objects():
        try:
            if torch.is_tensor(obj):
                if obj.is_cuda:
                    del obj
        except Exception:
            pass
    gc.collect()


def run_nvidia_smi() -> str:
    process = subprocess.Popen("nvidia-smi", shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    nvidia_smi_output, stderr = process.communicate()
    status = process.poll()
    if status != 0:
        print("Failed to run nvidia-smi")
        return None
    nvidia_smi_output = str(nvidia_smi_output)
    return nvidia_smi_output


def get_available_gpu_ids_from_nvidia_smi_output(raw_output: str) -> List[int]:
    """
    Parse 'nvidia-smi' table
    :param raw_output:
    :return:
    """
    split_output = raw_output.split("\\n")
    current_gpu = -1
    available_gpu_ids = []
    for line_idx, line in enumerate(split_output):
        if "Off" in line:
            current_gpu = int(split_output[line_idx][4])
            continue
        if current_gpu is not -1:
            loc = line.find("MiB /")
            if loc is not -1:
                memory_usage = int(line[int(loc - 7) : loc])
                if memory_usage < 50:
                    available_gpu_ids.append(current_gpu)
                    current_gpu = -1

    return available_gpu_ids



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""
from typing import Optional
from fuse.data.datasets.dataset_wrap_seq_to_dict import DatasetWrapSeqToDict
from torchvision import transforms

import torchvision

from fuse.data import DatasetDefault


class MNIST:
    """
    FuseMedML style of MNIST dataset: http://yann.lecun.com/exdb/mnist/
    """

    # bump whenever the static pipeline modified
    MNIST_DATASET_VER = 0

    @staticmethod
    def dataset(cache_dir: Optional[str] = None, train: bool = True) -> DatasetDefault:
        """
        Get mnist dataset - each sample includes: 'data.image', 'data.label' and 'data.sample_id'
        :param cache_dir: optional - destination to cache mnist
        :param train: If True, creates dataset from ``train-images-idx3-ubyte``,
            otherwise from ``t10k-images-idx3-ubyte``.
        """

        transform = transforms.Compose([transforms.ToTensor(), transforms.Normalize((0.1307,), (0.3081,))])
        # Create dataset
        torch_train_dataset = torchvision.datasets.MNIST(
            cache_dir, download=cache_dir is not None, train=train, transform=transform
        )
        # wrapping torch dataset
        train_dataset = DatasetWrapSeqToDict(
            name=f"mnist-{train}", dataset=torch_train_dataset, sample_keys=("data.image", "data.label")
        )
        train_dataset.create()
        return train_dataset



import paramiko
import re


class ShellHandler:
    def __init__(self, host, user, psw):
        self.ssh = paramiko.SSHClient()
        self.ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        self.ssh.connect(host, username=user, password=psw, port=22)

        channel = self.ssh.invoke_shell()
        self.stdin = channel.makefile("wb")
        self.stdout = channel.makefile("r")

    def __del__(self):
        self.ssh.close()

    def execute(self, cmd):
        """

        @param cmd: the command to be executed on the remote computer
        :examples:  execute('ls')
                    execute('finger')
                    execute('cd folder_name')
        """
        cmd = cmd.strip("\n")
        self.stdin.write(cmd + "\n")
        finish = "end of stdOUT buffer. finished with exit status"
        echo_cmd = "echo {} $?".format(finish)
        self.stdin.write(echo_cmd + "\n")
        shin = self.stdin
        self.stdin.flush()

        shout = []
        sherr = []
        exit_status = 0
        for line in self.stdout:
            if str(line).startswith(cmd) or str(line).startswith(echo_cmd):
                # up for now filled with shell junk from stdin
                shout = []
            elif str(line).startswith(finish):
                # our finish command ends with the exit status
                exit_status = int(str(line).rsplit(maxsplit=1)[1])
                if exit_status:
                    # stderr is combined with stdout.
                    # thus, swap sherr with shout in a case of failure.
                    sherr = shout
                    shout = []
                break
            else:
                # get rid of 'coloring and formatting' special characters
                shout.append(
                    re.compile(r"(\x9B|\x1B\[)[0-?]*[ -/]*[@-~]").sub("", line).replace("\b", "").replace("\r", "")
                )

        # first and last lines of shout/sherr contain a prompt
        if shout and echo_cmd in shout[-1]:
            shout.pop()
        if shout and cmd in shout[0]:
            shout.pop(0)
        if sherr and echo_cmd in sherr[-1]:
            sherr.pop()
        if sherr and cmd in sherr[0]:
            sherr.pop(0)

        return shin, shout, sherr



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from abc import ABC


class AlignMapBase(ABC):
    def __init__(self):
        """
        AlignMap settings, e.g. number of iterations for an iterative algorithm.
        """
        pass

    def align(self, img1, img2):
        """
        Learn mapping between two images. This may be a computationally heavy step.
        Mapping is unidirectional, from coordinates in img1 to coordinates in img2.
        Examples:
            For feature-based methods, calculate local features
            For deep learning methods, forward pass images
            For key-points based methods, find image key points (e.g. nipple)
        :param img1: ndarray, float32
        :param img2: ndarray, float32
        :return: None
        """
        pass

    def translate_xy(self, x, y):
        """
        Map coordinates: img1 --> img2
        Ideally this should be a computationally light step.
        :param x: horizontal position, measured from left side of images
        :param y: vertical position, measured from top of image
        :return: translated coordinates with respect to img2
        """
        pass



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from typing import Dict, Hashable, List, Optional, Sequence, Union

from warnings import warn
from fuse.data.datasets.dataset_base import DatasetBase
from fuse.data.ops.ops_common import OpCollectMarker
from fuse.data.pipelines.pipeline_default import PipelineDefault
from fuse.data.datasets.caching.samples_cacher import SamplesCacher
from fuse.utils.ndict import NDict
from fuse.utils.multiprocessing.run_multiprocessed import run_multiprocessed, get_from_global_storage
from fuse.data import get_sample_id, create_initial_sample, get_specific_sample_from_potentially_morphed
import copy
from collections import OrderedDict
import numpy as np


class DatasetDefault(DatasetBase):
    def __init__(
        self,
        sample_ids: Sequence[Hashable],
        static_pipeline: Optional[PipelineDefault] = None,
        dynamic_pipeline: Optional[PipelineDefault] = None,
        cacher: Optional[SamplesCacher] = None,
        allow_uncached_sample_morphing: bool = False,
    ):
        """
        :param sample_ids:      list of sample_ids included in dataset.
        :param static_pipeline:   static_pipeline, the output of this pipeline will be automatically cached.
        :param dynamic_pipeline:  dynamic_pipeline. applied sequentially after the static_pipeline, but not automatically cached.
                                changing it will NOT trigger recaching of the static_pipeline part.
        :param cacher: optional SamplesCacher instance which will be used for caching samples to speed up samples loading
        :param allow_uncached_sample_morphing:  when enabled, allows an Op, to return None, or to return multiple samples (in a list)

        """
        super().__init__()

        # store arguments
        self._static_pipeline = static_pipeline
        self._dynamic_pipeline = dynamic_pipeline
        self._cacher = cacher
        self._orig_sample_ids = sample_ids
        self._allow_uncached_sample_morphing = allow_uncached_sample_morphing

        # verify unique names for dynamic pipelines
        if self._dynamic_pipeline is not None and self._static_pipeline is not None:
            if self._static_pipeline.get_name() == self._dynamic_pipeline.get_name():
                raise Exception(
                    f"Detected identical name for static pipeline and dynamic pipeline ({self._static_pipeline.get_name(self._static_pipeline.get_name())}).\nThis is not allowed, please initiate the pipelines with different names."
                )

        if self._static_pipeline is None:
            self._static_pipeline = PipelineDefault("dummy_static_pipeline", ops_and_kwargs=[])
        if self._dynamic_pipeline is None:
            self._dynamic_pipeline = PipelineDefault("dummy_dynamic_pipeline", ops_and_kwargs=[])

        if self._dynamic_pipeline is not None:
            assert isinstance(
                self._dynamic_pipeline, PipelineDefault
            ), f"dynamic_pipeline may be None or a PipelineDefault instance. Instead got {type(self._dynamic_pipeline)}"

        if self._static_pipeline is not None:
            assert isinstance(
                self._static_pipeline, PipelineDefault
            ), f"static_pipeline may be None or a PipelineDefault instance. Instead got {type(self._static_pipeline)}"

        if self._allow_uncached_sample_morphing:
            warn(
                "allow_uncached_sample_morphing is enabled! It is a significantly slower mode and should be used ONLY for debugging"
            )

        self._created = False

    def create(self, num_workers: int = 0, mp_context: Optional[str] = None) -> None:
        """
        Create the data set, including caching
        :param num_workers: number of workers. used only when caching is disabled and allow_uncached_sample_morphing is enabled
            set num_workers=0 to disable multiprocessing (more convenient for debugging)
            Setting num_workers for caching is done in cacher constructor.
        :param mp_context: "fork", "spawn", "thread" or None for multiprocessing default
        :return: None
        """

        self._output_sample_ids_info = None
        if self._cacher is not None:
            self._output_sample_ids_info = self._cacher.cache_samples(self._orig_sample_ids)
        elif self._allow_uncached_sample_morphing:
            _output_sample_ids_info_list = run_multiprocessed(
                DatasetDefault._process_orig_sample_id,
                [(sid, self._static_pipeline, False) for sid in self._orig_sample_ids],
                workers=num_workers,
                mp_context=mp_context,
            )

            self._output_sample_ids_info = OrderedDict()
            self._final_sid_to_orig_sid = {}
            for sample_in_out_info in _output_sample_ids_info_list:
                orig_sid, out_sids = sample_in_out_info[0], sample_in_out_info[1]
                self._output_sample_ids_info[orig_sid] = out_sids
                if out_sids is not None:
                    assert isinstance(out_sids, list)
                    for final_sid in out_sids:
                        self._final_sid_to_orig_sid[final_sid] = orig_sid

        if self._output_sample_ids_info is not None:  # sample morphing is allowed
            self._final_sample_ids = []
            for orig_sid, out_sids in self._output_sample_ids_info.items():
                if out_sids is None:
                    continue
                self._final_sample_ids.extend(out_sids)
        else:
            self._final_sample_ids = copy.deepcopy(self._orig_sample_ids)

        self._created = True

    def get_all_sample_ids(self):
        if not self._created:
            raise Exception("you must first call create()")

        return copy.deepcopy(self._final_sample_ids)

    def __getitem__(self, item: Union[int, Hashable]) -> dict:
        """
        Get sample, read from cache if possible
        :param item: either int representing sample index or sample_id
        :return: sample_dict
        """
        return self.getitem(item)

    def getitem(
        self,
        item: Union[int, Hashable],
        collect_marker_name: Optional[str] = None,
        keys: Optional[Sequence[str]] = None,
    ) -> NDict:
        """
        Get sample, read from cache if possible
        :param item: either int representing sample index or sample_id
        :param collect_marker_name: Optional, specify name of collect marker op to optimize the running time
        :param keys: Optional, return just the specified keys or everything available if set to None
        :return: sample_dict
        """
        if not self._created:
            raise Exception("you must first call create()")

        # get sample id
        if isinstance(item, (int, np.integer)):
            sample_id = self._final_sample_ids[item]
        else:
            sample_id = item

        # get collect marker info
        collect_marker_info = self._get_collect_marker_info(collect_marker_name)

        # read sample
        if self._cacher is not None:
            sample = self._cacher.load_sample(sample_id, collect_marker_info["static_keys_deps"])

        if self._cacher is None:
            if not self._allow_uncached_sample_morphing:
                sample = create_initial_sample(sample_id)
                sample = self._static_pipeline(sample)
                if not isinstance(sample, dict):
                    raise Exception(
                        f'By default when caching is disabled sample morphing is not allowed, and the output of the static pipeline is expected to be a dict. Instead got {type(sample)}. You can use "allow_uncached_sample_morphing=True" to allow this, but be aware it is slow and should be used only for debugging'
                    )
            else:
                orig_sid = self._final_sid_to_orig_sid[sample_id]
                sample = create_initial_sample(orig_sid)
                sample = self._static_pipeline(sample)

                assert sample is not None
                sample = get_specific_sample_from_potentially_morphed(sample, sample_id)

        sample = self._dynamic_pipeline(sample, until_op_id=collect_marker_info["op_id"])

        if not isinstance(sample, dict):
            raise Exception(
                f"The final output of dataset static (+optional dynamic) pipelines is expected to be a dict. Instead got {type(sample)}"
            )

        # get just required keys
        if keys is not None:
            sample = sample.get_multi(keys)

        return sample

    def _get_multi_multiprocess_func(self, args):
        sid, kwargs = args
        return self.getitem(sid, **kwargs)

    @staticmethod
    def _getitem_multiprocess(item: Union[Hashable, int, np.integer]):
        """
        getitem method used to optimize the running time in a multiprocess mode
        """
        dataset = get_from_global_storage("dataset_default_get_multi_dataset")
        kwargs = get_from_global_storage("dataset_default_get_multi_kwargs")
        return dataset.getitem(item, **kwargs)

    def get_multi(
        self,
        items: Optional[Sequence[Union[int, Hashable]]] = None,
        workers: int = 10,
        verbose: int = 1,
        mp_context: Optional[str] = None,
        **kwargs,
    ) -> List[Dict]:
        """
        See super class
        :param workers: number of processes to read the data. set to 0 to not use multi processing (useful when debugging).
        :param mp_context: "fork", "spawn", "thread" or None for multiprocessing default
        """
        if items is None:
            sample_ids = self._final_sample_ids
        else:
            sample_ids = items

        for_global_storage = {"dataset_default_get_multi_dataset": self, "dataset_default_get_multi_kwargs": kwargs}

        list_sample_dict = run_multiprocessed(
            worker_func=self._getitem_multiprocess,
            copy_to_global_storage=for_global_storage,
            args_list=sample_ids,
            workers=workers,
            verbose=verbose,
            mp_context=mp_context,
        )
        return list_sample_dict

    def __len__(self):
        if not self._created:
            raise Exception("you must first call create()")

        return len(self._final_sample_ids)

    # internal methods

    @staticmethod
    def _process_orig_sample_id(args):
        """
        Process, without caching, single sample
        """
        orig_sample_id, pipeline, return_sample_dict = args
        sample = create_initial_sample(orig_sample_id)

        sample = pipeline(sample)

        output_sample_ids = None

        if sample is not None:
            output_sample_ids = []
            if not isinstance(sample, list):
                sample = [sample]
            for curr_sample in sample:
                output_sample_ids.append(get_sample_id(curr_sample))

        if not return_sample_dict:
            return orig_sample_id, output_sample_ids

        return orig_sample_id, output_sample_ids, sample

    def _get_collect_marker_info(self, collect_marker_name: str):
        """
        Find the required collect marker (OpCollectMarker in the dynamic pipeline).
        See OpCollectMarker for more details
        :param collect_marker_name: name to identify the required collect marker
        :return: a dictionary with the required info - including: name, op_id and static_keys_deps.
        if collect_marker_name is None will return default instruct to run the entire dynamic pipeline
        """
        # default values for case collect marker info is not used
        if collect_marker_name is None:
            return {"name": None, "op_id": None, "static_keys_deps": None}

        # find the required collect markers and extract the info
        collect_marker_info = None
        for (op, _), op_id in reversed(zip(self._dynamic_pipeline.ops_and_kwargs, self._dynamic_pipeline._op_ids)):
            if isinstance(op, OpCollectMarker):
                collect_marker_info_cur = op.get_info()
                if collect_marker_info_cur["name"] == collect_marker_name:
                    if collect_marker_info is None:
                        collect_marker_info = collect_marker_info_cur
                        collect_marker_info["op_id"] = op_id
                        # continue to make sure this is the only one
                    else:
                        # throw an error if found more than one collect marker
                        raise Exception(
                            f"Error: two collect markers with name {collect_marker_info} found in dynamic pipeline"
                        )
        if collect_marker_info is None:
            raise Exception(f"Error: didn't find collect marker with name {collect_marker_info} in dynamic pipeline.")

        return collect_marker_info

    def summary(self) -> str:
        sum = ""
        sum += f"Type: {type(self).__name__}\n"
        sum += f"Num samples: {len(self._final_sample_ids)}\n"
        # TODO
        # sum += f"Cacher: {self._cacher.summary()}"
        # sum += f"Pipeline static: {self._static_pipeline.summary()}"
        # sum += f"Pipeline dynamic: {self._dynamic_pipeline.summary()}"

        return sum



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from typing import Dict, Tuple, Sequence, Optional

import torch
import torch.nn as nn
import torch.nn.functional as F

from fuse.utils.ndict import NDict
from fuse.dl.models.heads.common import ClassifierFCN3D, ClassifierMLP


class Head3DClassifier(nn.Module):
    """
    Model that capture slice feature including the 3D context given the local feature about a slice for classification.
    """

    def __init__(
        self,
        head_name: str = "head_0",
        conv_inputs: Sequence[Tuple[str, int]] = None,
        dropout_rate: float = 0.1,
        num_classes: int = 3,
        append_features: Optional[Tuple[str, int]] = None,
        layers_description: Sequence[int] = (256,),
        append_layers_description: Sequence[int] = tuple(),
        append_dropout_rate: float = 0.0,
        fused_dropout_rate: float = 0.0,
    ) -> None:
        """
        Create simple 3D context model
        :param head_name: string representing the head name
        :param conv_inputs: Sequence of tuples, each indication features name in batch_dict and size of features (channels)
            for example: conv_inputs=(('model.backbone_features', 512),)
        :param dropout_rate: dropout fraction
        :param num_classes: number of output classes
        :param append_features: Sequence of tuples, each indication features name in batch_dict and size of features (channels).
                                Those are global features that appended after the global max pooling operation
        :param layers_description:          Layers description for the classifier module - sequence of hidden layers sizes (Not used currently)
        :param append_layers_description: Layers description for the tabular data, before the concatenation with the features extracted from the image - sequence of hidden layers sizes
        :param append_dropout_rate: Dropout rate for tabular layers
        """
        super().__init__()
        # save input params
        self.head_name = head_name
        self.conv_inputs = conv_inputs
        self.dropout_rate = dropout_rate
        self.num_classes = num_classes
        self.append_features = append_features
        self.gmp = nn.AdaptiveMaxPool3d(output_size=1)
        self.features_size = sum([features[1] for features in self.conv_inputs]) if self.conv_inputs is not None else 0

        # calc appended feature size if used
        if self.append_features is not None:
            if len(append_layers_description) == 0:
                self.features_size += sum([post_concat_input[1] for post_concat_input in append_features])
                self.append_features_module = nn.Identity()
            else:
                self.features_size += append_layers_description[-1]
                self.append_features_module = ClassifierMLP(
                    in_ch=sum([post_concat_input[1] for post_concat_input in append_features]),
                    num_classes=None,
                    layers_description=append_layers_description,
                    dropout_rate=append_dropout_rate,
                )

        self.conv_classifier_3d = ClassifierFCN3D(
            self.features_size, self.num_classes, layers_description, fused_dropout_rate
        )

        self.do = nn.Dropout3d(p=self.dropout_rate)

    def forward(self, batch_dict: NDict) -> Dict:
        """
        Forward pass
        :param batch_dict: dictionary containing an input tensor representing spatial features with 3D context. shape: [batch_size, in_features, z, y, x]
        :return: batch dict with fields model.outputs and model.logits
        """
        if self.conv_inputs is not None:
            conv_input = torch.cat([batch_dict[conv_input[0]] for conv_input in self.conv_inputs], dim=1)
            global_features = self.gmp(conv_input)
            # backward compatibility
            if hasattr(self, "do"):
                global_features = self.do(global_features)
        # append global features if are used
        if self.append_features is not None:
            features = torch.cat(
                [batch_dict[features[0]].reshape(-1, features[1]) for features in self.append_features], dim=1
            )
            features = self.append_features_module(features)
            features = features.reshape(features.shape + (1, 1, 1))
            if self.conv_inputs is not None:
                global_features = torch.cat((global_features, features), dim=1)
            else:
                global_features = features

        logits = self.conv_classifier_3d(global_features)
        logits = logits.squeeze(dim=4)
        logits = logits.squeeze(dim=3)
        logits = logits.squeeze(dim=2)  # squeeze will change the shape to  [batch_size, channels']

        cls_preds = F.softmax(logits, dim=1)
        batch_dict["model.logits." + self.head_name] = logits
        batch_dict["model.output." + self.head_name] = cls_preds

        return batch_dict


class Head3DRegression(nn.Module):
    """
    Model that capture slice feature including the 3D context given the local feature about a slice for regression.
    """

    def __init__(
        self,
        head_name: str = "head_0",
        conv_inputs: Sequence[Tuple[str, int]] = None,
        dropout_rate: float = 0.1,
        append_features: Optional[Tuple[str, int]] = None,
        append_layers_description: Sequence[int] = tuple(),
        append_dropout_rate: float = 0.0,
        num_outputs: int = 1,
    ) -> None:
        """
        Create simple 3D context model
        :param head_name: string representing the head name
        :param conv_inputs: Sequence of tuples, each indication features name in batch_dict and size of features (channels)
            for example: conv_inputs=(('model.backbone_features', 512),)
        :param dropout_rate: dropout fraction
        :param append_features: Sequence of tuples, each indication features name in batch_dict and size of features (channels).
                                Those are global features that appended after the global max pooling operation
        :param layers_description:          Layers description for the classifier module - sequence of hidden layers sizes (Not used currently)
        :param append_layers_description: Layers description for the tabular data, before the concatenation with the features extracted from the image - sequence of hidden layers sizes
        :param append_dropout_rate: Dropout rate for tabular layers
        :param num_outputs: number of output values
        """
        super().__init__()
        # save input params
        self.head_name = head_name
        self.conv_inputs = conv_inputs
        self.dropout_rate = dropout_rate
        self.append_features = append_features
        self.gmp = nn.AdaptiveMaxPool3d(output_size=1)
        self.features_size = sum([features[1] for features in self.conv_inputs]) if self.conv_inputs is not None else 0

        # calc appended feature size if used
        if self.append_features is not None:
            if len(append_layers_description) == 0:
                self.features_size += sum([post_concat_input[1] for post_concat_input in append_features])
                self.append_features_module = nn.Identity()
            else:
                self.features_size += append_layers_description[-1]
                self.append_features_module = ClassifierMLP(
                    in_ch=sum([post_concat_input[1] for post_concat_input in append_features]),
                    num_classes=None,
                    layers_description=append_layers_description,
                    dropout_rate=append_dropout_rate,
                )

        self.linear = nn.Linear(self.features_size, num_outputs)

        self.do = nn.Dropout3d(p=self.dropout_rate)

    def forward(self, batch_dict: NDict) -> Dict:
        """
        Forward pass
        :param batch_dict: dictionary containing an input tensor representing spatial features with 3D context. shape: [batch_size, in_features, z, y, x]
        :return: batch dict with fields model.outputs and model.logits
        """
        if self.conv_inputs is not None:
            conv_input = torch.cat([batch_dict[conv_input[0]] for conv_input in self.conv_inputs], dim=1)
            global_features = self.gmp(conv_input)
            # backward compatibility
            if hasattr(self, "do"):
                global_features = self.do(global_features)
        # append global features if are used
        if self.append_features is not None:
            features = torch.cat(
                [batch_dict[features[0]].reshape(-1, features[1]) for features in self.append_features], dim=1
            )
            features = self.append_features_module(features)
            features = features.reshape(features.shape + (1, 1, 1))
            if self.conv_inputs is not None:
                global_features = torch.cat((global_features, features), dim=1)
            else:
                global_features = features
        global_features = global_features.squeeze(dim=4)
        global_features = global_features.squeeze(dim=3)
        global_features = global_features.squeeze(dim=2)
        prediction = self.linear(global_features).squeeze(dim=1)
        batch_dict["model.output." + self.head_name] = prediction

        return batch_dict



from collections import OrderedDict
from typing import Any, Sequence
import re


class Patterns:
    """
    Utility to match a string to a pattern.
    Typically used to infer data type from key in sample_dict
    """

    def __init__(self, patterns_dict: OrderedDict, default_value: Any = None):
        """
        :param patterns_dict: ordered dictionary, the key is a regex expression.
                        The value of the first matched key will be returned.
        Example:
        patterns = {
            r".*img$": DataType.IMAGE,
            r".*seg$": DataType.SEG,
            r".*bbox$": DataType.BBOX,
            r".*$": DataType.UNKNOWN
        }
        pp = Patterns(patterns)
        print(pp.get_type("data.cc.img")) -> DataType.IMAGE
        print(pp.get_type("data.cc_img")) -> DataType.IMAGE
        print(pp.get_type("data.img_seg")) -> DataType.SEG
        print(pp.get_type("data.imgseg")) -> DataType.SEG
        print(pp.get_type("data")) -> DataType.UNKNOWN
        print(pp.get_type("bbox")) -> DataType.BBox
        print(pp.get_type("a.bbox")) -> DataType.BBOX

        :param default_value: value to return in case there is not match
        """
        self._patterns = patterns_dict
        self._default_value = default_value

    def get_value(self, key: str) -> Any:
        """
        :param key: string to match
        :return: the first value from patterns with pattern that match to key
        """
        for pattern in self._patterns:
            if re.match(pattern, key) is not None:
                return self._patterns[pattern]

        return self._default_value

    def verify_value_in(self, key: str, values: Sequence[Any]) -> None:
        """
        Raise an exception of the matched value not in values
        :param key: string to match
        :param values: list of supported values
        :return: None
        """
        val_type = self.get_value(key)
        if val_type not in values:
            raise ValueError(
                f"key {key} mapped to unsupported type {val_type}.\n List of supported types {values} \n Patterns {self._patterns}"
            )



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, Hashable, Optional, Sequence, Tuple, Union
import copy
from fuse.utils import uncollate

import pandas as pd
import torch
import numpy as np
import scipy

from fuse.utils import NDict


class MetricBase(ABC):
    """
    Required interface for a metric implementation
    """

    @abstractmethod
    def collect(self, batch: Dict) -> None:
        """
        aggregate data from batchs
        :param batch: bath dict
        """
        raise NotImplementedError

    @abstractmethod
    def set(self, data: pd.DataFrame) -> None:
        """
        set entire data at once
        :param data: dataframe representation of the data. Each line is a sample and each column is a value field.
        """
        raise NotImplementedError

    @abstractmethod
    def reset(self) -> None:
        """
        reset state
        """
        raise NotImplementedError

    @abstractmethod
    def eval(
        self, results: Dict[str, Any] = None, ids: Optional[Sequence[Hashable]] = None
    ) -> Union[Dict[str, Any], Any]:
        """
        evaluate the collected data
        :param results: results aggergated by the previous metrics
        :param ids: sequence of sample ids
        The evaluated results should be stored in results. (recommendation: store it under metrics namespace)
        """
        raise NotImplementedError


class MetricCollector(MetricBase):
    """
    Collect data for metrics with native support for data sampling
    """

    DEFAULT_ID_KEYS = ["id", "data.sample_id", "sample_id"]

    def __init__(
        self,
        pre_collect_process_func: Optional[Callable] = None,
        post_collect_process_func: Optional[Callable] = None,
        **keys_to_collect: Dict[str, str],
    ):
        """
        :param pre_collect_process_func: Optional callable - the callable will get as an input a sample_dict and can preprocess it if required
        :param post_collect_process_func: Optional callable - custom process func that convert the fields to be collected to the values that will actually be collected
                                          the callable will get as an input the collected values of a single sample
                                          and can return either a single value or dictionary. The returned values will be collected under the name "post_args"
        :param keys_to_collect: specify the keys you want to collect from the source data
        """
        super().__init__()
        # store input
        self._pre_collect_process_func = pre_collect_process_func
        self._post_collect_process_func = post_collect_process_func
        self._keys_to_collect = copy.copy(keys_to_collect)
        self._id_keys = MetricCollector.DEFAULT_ID_KEYS

        # reset
        self.reset()

    def collect(self, batch: Dict) -> None:
        """
        See super class
        """
        if not isinstance(batch, NDict):
            batch = NDict(batch)

        samples = uncollate(batch)
        for sample in samples:
            sample_to_collect = {}

            if self._pre_collect_process_func is not None:
                sample = NDict(self._pre_collect_process_func(sample))

            for name, key in self._keys_to_collect.items():
                value = sample[key]
                if isinstance(value, torch.Tensor):
                    value = value.detach().cpu().numpy()

                sample_to_collect[name] = value

            if self._post_collect_process_func is not None:
                sample_to_collect = {"post_args": self._post_collect_process_func(**sample_to_collect)}

            # store it - assumes batch dimension? What about single sample?
            for name in sample_to_collect:
                self._collected_data[name].append(sample_to_collect[name])

        # extract ids and store it in self._collected_ids
        ids = None
        for key in self._id_keys:
            if key in batch:
                ids = batch[key]
                break

        if ids is not None:
            self._collected_ids.extend(ids)

    @staticmethod
    def _df_dict_apply(data: pd.Series, func: Callable) -> pd.Series:
        result = func(NDict(data.to_dict()))
        return pd.Series(result.flatten())

    @staticmethod
    def _df_dict_apply_kwargs(data: pd.Series, func: Callable) -> pd.Series:
        result = func(**data.to_dict())
        return pd.Series(result.flatten())

    def set(self, data: pd.DataFrame) -> None:
        """
        see super class
        """
        self.reset()

        if self._pre_collect_process_func is not None:
            pre_collect_process = lambda x: self._df_dict_apply(x, self._pre_collect_process_func)
            data = data.apply(pre_collect_process, axis=1)

        data_to_collect = pd.DataFrame(data=None, columns=self._keys_to_collect)
        for name, key in self._keys_to_collect.items():
            if key not in data.keys():
                raise Exception(f"Error key {key} wasn't found. Available keys {data.keys()}")

            data_to_collect[name] = data[key]

        if self._post_collect_process_func is not None:
            post_collect_process = lambda x: self._df_dict_apply_kwargs(x, self._post_collect_process_func)
            data_to_collect = data.apply(post_collect_process, axis=1)

        for name in data_to_collect.keys():
            values = data_to_collect.loc[:, name]
            self._collected_data[name].extend(values)

        # extract ids and store it in self._collected_ids
        ids = None
        for key in self._id_keys:
            if key in data.keys():
                ids = list(data[key])
                break

        if ids is not None:
            self._collected_ids.extend(ids)

    def reset(self) -> None:
        """
        See super class
        """
        if self._post_collect_process_func is None:
            self._collected_data = {name: [] for name in self._keys_to_collect}
        else:
            # collect everything you get from post_collect_process_args
            self._collected_data = {"post_args": []}

        self._collected_ids = []  # the original collected ids

        self._sampled_ids = None  # the required ids - set be sample() method

    def get_ids(self) -> Sequence[Hashable]:
        """
        See super class
        """
        return self._collected_ids

    def get(self, ids: Optional[Sequence[Hashable]] = None) -> Tuple[Dict[str, Any]]:
        """
        Get collected data - collected data dictionary and collected ids.
        each element in the dictionary will be a list of values from all samples
        """
        if ids is None:
            return copy.copy(self._collected_data)
        else:
            # convert required ids to permutation
            original_ids = self._collected_ids
            required_ids = ids
            permutation = [original_ids.index(sample_id) for sample_id in required_ids]

            # create the permuted dictionary
            data = {}
            for name, values in self._collected_data.items():
                data[name] = [values[i] for i in permutation]

            return data

    def eval(self, results: Dict[str, Any] = None) -> Union[Dict[str, Any], Any]:
        """
        Empty implementation - do nothing
        """
        pass


class MetricWithCollectorBase(MetricBase):
    """
    Base implementation of metric with built-in collector
    There is also an option for an external data collector  - in a case space optimization required and there by using shared collector for few metrics
    """

    def __init__(
        self,
        pre_collect_process_func: Optional[Callable] = None,
        post_collect_process_func: Optional[Callable] = None,
        external_data_collector: Optional[MetricCollector] = None,
        extract_ids: bool = False,
        **kwargs,
    ) -> None:
        """
        :param pre_collect_process_func: Optional callable - the callable will get as an input a batch_dict or a dataframe and can preprocess it if required
        :param external_data_collector: Optional - in a case space optimization required and there by using shared collector for few metrics
        :param extract_ids: self._extract_arguments packs all arguments for a underlying function. Set to True, to pack also the ids (under the name 'ids')
        :param kwargs: specify keyworded and value arguments you want to collect from the source data.
                can be strings (key names) and/or actual values
                to collect from the results dictionary: add a "results:" prefix to the key name
        """
        super().__init__()
        self._keys_to_collect = {n: k for n, k in kwargs.items() if isinstance(k, str) and not k.startswith("results:")}
        self._keys_from_results = {
            n: k[len("results:") :] for n, k in kwargs.items() if isinstance(k, str) and k.startswith("results:")
        }
        self._value_args = {n: k for n, k in kwargs.items() if k is not None and not isinstance(k, str)}
        self._collector = (
            MetricCollector(pre_collect_process_func, post_collect_process_func, **self._keys_to_collect)
            if external_data_collector is None
            else external_data_collector
        )
        self._collect_data_flag = external_data_collector is None
        self._extract_ids = extract_ids

    def collect(self, batch: Dict) -> None:
        """
        See super class
        """
        if self._collect_data_flag:
            return self._collector.collect(batch)

    def set(self, data: pd.DataFrame) -> None:
        """
        See super class
        """
        if self._collect_data_flag:
            return self._collector.set(data)

    def reset(self) -> None:
        """
        See super class
        """
        if self._collect_data_flag:
            return self._collector.reset()

    def _extract_arguments(self, results: Dict[str, Any] = None, ids: Optional[Sequence[Hashable]] = None) -> Dict:
        """
        extract keyworded arguments and value arguments from collected data and results dict
        """
        if not isinstance(results, NDict):
            results = NDict(results)

        arg_dict = {}
        data = self._collector.get(ids)
        arg_dict.update(data)

        for name in self._keys_from_results:
            res = results[self._keys_from_results[name]]
            if isinstance(res, Callable):  # per-sample metric
                res = res(ids)
            arg_dict[name] = res

        for name in self._value_args:
            arg_dict[name] = self._value_args[name]

        # pack also ids if those are required - see details in self.__init__
        if self._extract_ids:
            if ids is None:
                arg_dict["ids"] = self._collector.get_ids()
            else:
                arg_dict["ids"] = ids

        return arg_dict

    @abstractmethod
    def eval(self, results: Dict[str, Any] = None, ids: Optional[Sequence[Hashable]] = None) -> None:
        """
        See super class
        """
        raise NotImplementedError


class MetricDefault(MetricWithCollectorBase):
    """
    Default generic implementation for metric
    Can be used for any metric getting as an input list of prediction, list of targets and optionally additional parameters
    """

    def __init__(self, metric_func: Callable, pred: Optional[str] = None, target: Optional[str] = None, **kwargs):
        """
        :param pred: prediction key to collect
        :param target: target key to collect
        :param metric_func: function getting as a input list of predictions, targets and optionally more arguments specified in kwargs
                            the function should return a single result or a dictionary of results
        :param kwargs: additional keyword arguments for MetricWithCollectorBase.
                       The keyword expected to be an argument name of metric_func and the value a string that is a key to value store in batch dict.
                       If instead a value should be extracted from results dict use "results:<key in results dict>
                       Example:
                       def my_metric_func(pred, target, sample_weight, operation_point):
                           ...
                       metric = MetricDefault(metric_func=my_metric_func,
                                              pred="model.classification.preds", target="data.classification.labels",
                                              sample_weight="data.classification.weight", operation_point="results:metrics.operation_point")
        """
        super().__init__(pred=pred, target=target, **kwargs)
        self._metric_func = metric_func

    def eval(self, results: Dict[str, Any] = None, ids: Optional[Sequence[Hashable]] = None) -> Union[Dict, Any]:
        """
        See super class
        """

        # extract values from collected data and results dict
        kwargs = self._extract_arguments(results, ids)

        # single evaluation method
        return self._metric_func(**kwargs)


class MetricPerSampleDefault(MetricWithCollectorBase):
    """
    Default generic implementation for a case that is better to compute the metric per sample and then aggregate the results.
    Can be used for any metric getting as an input list of prediction, list of targets and optionally additional parameters
    """

    def __init__(
        self, pred: str, target: str, metric_per_sample_func: Callable, result_aggregate_func: Callable, **kwargs
    ):
        """
        :param pred: prediction key to collect
        :param target: target key to collect
        :param metric_per_sample_func: function that gets as input the values to collect specified in metric constructor, typically pred and target.
                                       A sequence of all the returned values from all samples will be passed to result_aggregate_func
        :param result_aggregate_func: function that get the output of metric_per_sample_func and aggregates it over multiple samples
        :param kwargs: additional kw arguments for MetricWithCollectorBase
        """

        super().__init__(pred=pred, target=target, post_collect_process_func=metric_per_sample_func, **kwargs)
        self._result_aggregate_func = result_aggregate_func

    def eval(self, results: Dict[str, Any] = None, ids: Optional[Sequence[Hashable]] = None) -> Union[Dict, Any]:
        """
        See super class
        """

        # extract values from collected data and results dict
        kwargs = self._extract_arguments(results, ids)

        # single evaluation method
        return self._result_aggregate_func(kwargs["post_args"])


class GroupAnalysis(MetricWithCollectorBase):
    """
    Evaluate a metric per group and compute basic statistics about the different per group results.
    eval() method returns a dictionary of the following format:
    {'mean': <>, 'std': <>, 'median': <>, <group 0>: <>, <group 1>: <>, ...}
    """

    def __init__(self, metric: MetricBase, group: str, **super_kwargs) -> None:
        """
        :param metric: metric to analyze
        :param group: key to extract the group from
        :param super_kwargs: additional arguments for super class (MetricWithCollectorBase) constructor
        """
        super().__init__(group=group, **super_kwargs)
        self._metric = metric

    def collect(self, batch: Dict) -> None:
        "See super class"
        self._metric.collect(batch)
        return super().collect(batch)

    def set(self, data: pd.DataFrame) -> None:
        "See super class"
        self._metric.set(data)
        return super().set(data)

    def reset(self) -> None:
        "See super class"
        self._metric.reset()
        return super().reset()

    def eval(self, results: Dict[str, Any] = None, ids: Optional[Sequence[Hashable]] = None) -> Dict[str, Any]:
        """
        See super class
        :return: a dictionary of the following format {'mean': <>, 'std': <>, 'median': <>, <group 0>: <>, <group 1>: <>, ...}
        """
        data = self._collector.get(ids)
        if ids is None:
            ids = self._collector.get_ids()
        if not ids:
            raise Exception(
                "Error: group analysis is supported only when a unique identifier is specified. Add key 'id' to your data"
            )
        ids = np.array(ids)

        groups = np.array(data["group"])
        unique_groups = set(groups)

        group_analysis_results = {}
        for group_value in unique_groups:
            group_ids = ids[groups == group_value]

            group_analysis_results[str(group_value)] = self._metric.eval(results, group_ids)

        # compute stats
        group_results_list = list(group_analysis_results.values())
        if isinstance(group_results_list[0], dict):  # multiple values
            # get all keys
            all_keys = set()
            for group_result in group_results_list:
                all_keys |= set(group_result.keys())

            for key in all_keys:
                values = [group_result[key] for group_result in group_results_list]
                try:
                    group_analysis_results[f"{key}.mean"] = np.mean(values)
                    group_analysis_results[f"{key}.std"] = np.std(values)
                    group_analysis_results[f"{key}.median"] = np.median(values)
                except:
                    # do nothing
                    pass
        else:  # single value
            values = [group_result for group_result in group_results_list]
            try:
                group_analysis_results["mean"] = np.mean(values)
                group_analysis_results["std"] = np.std(values)
                group_analysis_results["median"] = np.median(values)
            except:
                # do nothing
                pass

        return group_analysis_results


class Filter(MetricWithCollectorBase):
    """
    Evaluate a sub-group of data. This utility will filter non relevant samples and will call to the given metric.
    """

    def __init__(self, metric: MetricBase, filter: str, **super_kwargs) -> None:
        """
        :param metric: metric to filter samples for
        :param group: key to extract filter
        :param super_kwargs: additional arguments for super class (MetricWithCollectorBase) constructor
        """
        super().__init__(filter=filter, **super_kwargs)
        self._metric = metric

    def collect(self, batch: Dict) -> None:
        "See super class"
        self._metric.collect(batch)
        return super().collect(batch)

    def set(self, data: pd.DataFrame) -> None:
        "See super class"
        self._metric.set(data)
        return super().set(data)

    def reset(self) -> None:
        "See super class"
        self._metric.reset()
        return super().reset()

    def eval(self, results: Dict[str, Any] = None, ids: Optional[Sequence[Hashable]] = None) -> Dict[str, Any]:
        """
        See super class
        :return: a dictionary of the following format {'mean': <>, 'std': <>, 'median': <>, <group 0>: <>, <group 1>: <>, ...}
        """
        data = self._collector.get()
        if ids is None:
            ids = self._collector.get_ids()
        if not ids:
            raise Exception(
                "Error: group analysis is supported only when a unique identifier is specified. Add key 'id' to your data"
            )
        ids = np.array(ids)

        filter = np.array(data["filter"])
        keep_ids = ids[~filter]
        metric_result = self._metric.eval(results, keep_ids)
        return metric_result


class CI(MetricWithCollectorBase):
    """
    Compute confidence interval for a metric
    eval() method returns a dictionary of the following format:
    {'org': <>, 'mean': <>, 'std': <>, 'conf_interval': <>, 'conf_lower': <>, 'conf_upper': <>}
    """

    def __init__(
        self,
        metric: MetricBase,
        stratum: str,
        num_of_bootstraps: int = 10000,
        rnd_seed: int = 1234,
        conf_interval: float = 95,
        ci_method: str = "PERCENTILE",
        **super_kwargs,
    ) -> None:
        """
        :param metric: metric to compute the confidence interval for
        :param stratum: if sampling should be done by strata, specify the key in batch_dict to collect data from
        :param num_of_bootstraps: number of bootstrap iterations
        :param rnd_seed: seed for random number generator.
        :param conf_interval: Confidence interval. Default is 95.
        :param ci_method: specifies the method for computing the confidence intervals from bootstrap samples. Options: NORMAL (assuming normal distribution), PERCENTILE, PIVOTAL
        """
        super().__init__(stratum=stratum, **super_kwargs)

        self._metric = metric
        self._num_of_bootstraps = num_of_bootstraps
        self._rnd_seed = rnd_seed
        self._conf_interval = conf_interval
        self._ci_method = ci_method

        # verify
        assert ci_method in [
            "NORMAL",
            "PERCENTILE",
            "PIVOTAL",
        ], f"Error: unexpected confidence interval method: {ci_method}"

    def collect(self, batch: Dict) -> None:
        "See super class"
        self._metric.collect(batch)
        return super().collect(batch)

    def set(self, data: pd.DataFrame) -> None:
        "See super class"
        self._metric.set(data)
        return super().set(data)

    def reset(self) -> None:
        "See super class"
        self._metric.reset()
        return super().reset()

    def eval(self, results: Dict[str, Any] = None, ids: Optional[Sequence[Hashable]] = None) -> Dict[str, Any]:
        """
        See super class
        :return: dictionary of format - {'org': <>, 'mean': <>, 'std': <>, 'conf_interval': <>, 'conf_lower': <>, 'conf_upper': <>}
        """
        data = self._collector.get()
        if ids is None:
            ids = self._collector.get_ids()
        if not ids:
            raise Exception(
                "Error: confidence interval is supported only when a unique identifier is specified. Add key 'id' to your data"
            )
        ids = np.array(ids)

        rnd = np.random.RandomState(self._rnd_seed)
        original_sample_results = self._metric.eval(results)
        boot_results = []
        ci_results = {}

        stratum_id = np.array(data["stratum"]) if "stratum" in data else np.ones(len(data["index"]))
        unique_strata = np.unique(stratum_id)

        for _ in range(self._num_of_bootstraps):
            sampled_ids = ids.copy()
            for stratum in unique_strata:
                stratum_filter = stratum_id == stratum
                n_stratum = sum(stratum_filter)
                random_sample = rnd.randint(0, n_stratum, size=n_stratum)
                sampled_ids[stratum_filter] = ids[stratum_filter][random_sample]
            boot_results.append(self._metric.eval(results, sampled_ids))

        # results can be either a list of floats or a list of dictionaries
        if isinstance(original_sample_results, dict):
            for key, orig_val in original_sample_results.items():
                sampled_vals = [sample[key] for sample in boot_results]
                try:
                    ci_results[key] = self._compute_stats(self._ci_method, orig_val, sampled_vals, self._conf_interval)
                except:
                    ci_results[key] = orig_val
        elif isinstance(original_sample_results, float):
            ci_results = self._compute_stats(
                self._ci_method, original_sample_results, boot_results, self._conf_interval
            )

        return ci_results

    @staticmethod
    def _compute_stats(
        ci_method: str, orig: Union[float, np.ndarray], samples: Sequence[Union[float, np.ndarray]], confidence: float
    ) -> Dict[str, Union[np.ndarray, float]]:
        "Compute and package into dictionary CI results"
        confidence_lower, confidence_upper = CI.compute_confidence_interval(ci_method, orig, samples, confidence)
        return {
            "org": orig,
            "mean": np.mean(samples),
            "std": np.std(samples),
            "conf_interval": confidence,
            "conf_lower": confidence_lower,
            "conf_upper": confidence_upper,
        }

    @staticmethod
    def compute_confidence_interval(
        method: str,
        org_statistic: Union[float, np.ndarray],
        bootstrap_statistics: Sequence[Union[float, np.ndarray]],
        confidence: float,
    ) -> Tuple[float, float]:
        lower_confidence_ratio = (1 - confidence / 100.0) / 2.0
        if method == "NORMAL":
            sigma = np.std(bootstrap_statistics)
            c_alpha = scipy.stats.norm.isf(lower_confidence_ratio)
            return org_statistic - c_alpha * sigma, org_statistic + c_alpha * sigma

        statistic_observations = np.sort(bootstrap_statistics)

        n = len(statistic_observations)
        index_low = int(np.floor(lower_confidence_ratio * n))
        index_high = int(np.floor((1 - lower_confidence_ratio) * n))

        bootstrap_statistic_low = statistic_observations[index_low]
        bootstrap_statistic_high = statistic_observations[index_high]
        if method == "PERCENTILE":
            return bootstrap_statistic_low, bootstrap_statistic_high

        assert method == "PIVOTAL"
        return 2 * org_statistic - bootstrap_statistic_high, 2 * org_statistic - bootstrap_statistic_low



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from fuse.utils.align.utils_align_base import AlignMapBase
import numpy as np
import cv2


class AlignMapECC(AlignMapBase):
    def __init__(self, transformation_type="homography", num_iterations=600, termination_eps=1e-4):
        transformation_type = transformation_type.lower()
        assert transformation_type in ["homography", "affine"]

        self.transformation_type = transformation_type.lower()
        self.num_iterations = num_iterations
        self.termination_eps = termination_eps
        self.transformation_matrix = None
        self.img1_aligned = None
        self.img2_aligned = None

    def align(self, img1, img2, mask=None):

        if self.transformation_type == "affine":
            transformation = cv2.MOTION_AFFINE
            M = np.eye(2, 3, dtype=np.float32)
        elif self.transformation_type == "homography":
            transformation = cv2.MOTION_HOMOGRAPHY
            M = np.eye(3, 3, dtype=np.float32)
        else:
            raise Exception(
                f"Unknown transformation {self.transformation_type}. Allowed values are : ['affine', 'homography']"
            )

        # convert from [0, 1] to binary uint8
        img1_original = img1.copy()
        img2_original = img2.copy()
        img1 = ((img1 > 0) * 255).astype(np.uint8)
        img2 = ((img2 > 0) * 255).astype(np.uint8)

        criteria = (cv2.TERM_CRITERIA_EPS | cv2.TERM_CRITERIA_COUNT, self.num_iterations, self.termination_eps)

        s, M = cv2.findTransformECC(img1, img2, M, transformation, criteria=criteria, inputMask=mask)

        self.transformation_matrix = M

        # Apply the alignment on img1 and the inverse alignment on img2
        if transformation == cv2.MOTION_AFFINE:
            self.img1_aligned = cv2.warpAffine(
                img1_original, self.transformation_matrix, (img2.shape[1], img2.shape[0]), flags=cv2.INTER_LINEAR
            )
            self.img2_aligned = cv2.warpAffine(
                img2_original,
                self.transformation_matrix,
                (img1.shape[1], img1.shape[0]),
                flags=cv2.INTER_LINEAR + cv2.WARP_INVERSE_MAP,
            )
        else:
            self.img1_aligned = cv2.warpPerspective(
                img1_original, self.transformation_matrix, (img2.shape[1], img2.shape[0]), flags=cv2.INTER_LINEAR
            )
            self.img2_aligned = cv2.warpPerspective(
                img2_original,
                self.transformation_matrix,
                (img1.shape[1], img1.shape[0]),
                flags=cv2.INTER_LINEAR + cv2.WARP_INVERSE_MAP,
            )

        self.mask = mask

    def translate_xy(self, x, y):
        if self.transformation_matrix is None:
            raise Exception("No transformation matrix found - need to run 'align' first!")

        if self.transformation_type == "homography":
            wx = (
                self.transformation_matrix[0, 0] * x
                + self.transformation_matrix[0, 1] * y
                + self.transformation_matrix[0, 2]
            ) / (
                self.transformation_matrix[2, 0] * x
                + self.transformation_matrix[2, 1] * y
                + self.transformation_matrix[2, 2]
            )
            wy = (
                self.transformation_matrix[1, 0] * x
                + self.transformation_matrix[1, 1] * y
                + self.transformation_matrix[1, 2]
            ) / (
                self.transformation_matrix[2, 0] * x
                + self.transformation_matrix[2, 1] * y
                + self.transformation_matrix[2, 2]
            )
        else:
            raise Exception("Currently supports only homography translation")

        return wx, wy
